/*

              ####### #         ###   #     #   ###   ######
              #       #          #     #   #     #    #     #
              #       #          #      # #      #    #     #
              #####   #          #       #       #    ######
              #       #          #      # #      #    #   #
              #       #          #     #   #     #    #    #
              ####### #######   ###   #     #   ###   #     #

                   
                  Copyright: 1994 Petr Krysl

   Czech Technical University in Prague, Faculty of Civil Engineering,
      Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                  email: pk@power2.fsv.cvut.cz
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "Esimple.h"          


static int
has_suffix (char *f, char *s /* suffix */)
{
  int lf, ls, i;

  if (!f || !s)
    return FALSE;
  
  lf = strlen((f)); ls = min(strlen((s)), lf);
  for (i = 0; i < ls; i++)
    if (toupper((f)[lf-i]) != toupper((s)[ls-i]))
      return FALSE;
  return TRUE;
}


static void
suck_in_file (char *f)
{
  char command[1024];

  sprintf (command, "LOAD \"%s\"", f);
  ESIHandleCmd (command);
}


static int 
vrml_yourself (NODE ptr, NODE elem)
{
  EXVRMLWriteGeom ((FILE *)ptr, ESIModel (), (GraphicObj *)elem);
  return 1;
}

static void
write_wrl (char *f)
{
  GraphicObj *g_p;
  FILE *fp;
  fp = fopen (f, "w");
  if (!fp) {
    fprintf (stderr, "Could not write \"%s\"\n", f);
    exit (1);
  }
  EXVRMLStart (ESIModel (), fp);
  EMDispatchToGraphicObjects (ESIModel (), vrml_yourself, (caddr_t) fp);
  EXVRMLClose (fp);
  fclose(fp);
}


static char *
wrl_name (char *f)
{
  int tocopy;
  char *wrl;
  char ext[] = ".wrl";
  int i; 

  wrl = malloc (strlen (f) + strlen (ext) + 1);
  tocopy = strlen (f);
  if (has_suffix (f, ".egf")) {
    tocopy -= 4;
  }
  for (i = 0; i < tocopy; i++) wrl[i] = f[i];
  for (i = 0; i < 4; i++) wrl[tocopy + i] = ext[i];
  wrl[tocopy + i] = '\0';
  return wrl;
}


static void
process_args (int argc, char **argv)
{
  int i;
  int wrote = FALSE;
  char *last = NULL;
  
  i = 1;
  while (i < argc) {
    if        (has_suffix (argv[i], ".egf")) {
      if (ckit_io_file_exists (argv[i])) {
        suck_in_file (argv[i]);
        if (last) free  (last);
        last = strdup (argv[i]);
      }
    } else if (has_suffix (argv[i], ".wrl")) {
      write_wrl (argv[i]);
      wrote = TRUE;
    } else {
      char *withsuf;
      withsuf = malloc (strlen (argv[i]) + strlen (".egf") + 1);
      sprintf (withsuf, "%s%s", argv[i], ".egf");
      if (ckit_io_file_exists (withsuf)) {
        suck_in_file (withsuf);
        if (last) free  (last);
        last = strdup (withsuf);
      }
      free (withsuf);
    }
    i++;
  }
  if (!wrote) {
    char *wrl;
    if (last) {
      wrl = wrl_name (last);
      write_wrl (wrl);
      free (wrl);
    } else {
      fprintf (stderr, "No input?\n");
      exit (1);
    }
  }
}

int  
main(int argc, char **argv)
{
  unsigned long m = 0;
  m |= ESI_GRAPHIC_EDITOR_MASK;
  ESIBuildInterface(m, argc, argv);
  ESIHandleCmd("COLOR yellow");
  ESIHandleCmd("SHRINK 0.9");
  process_args (argc, argv);
  return 1;
}


 



void
ESICustomize(Widget parent_pane)
{
/* VOID */
}
