/*

              ####### #         ###   #     #   ###   ######
              #       #          #     #   #     #    #     #
              #       #          #      # #      #    #     #
              #####   #          #       #       #    ######
              #       #          #      # #      #    #   #
              #       #          #     #   #     #    #    #
              ####### #######   ###   #     #   ###   #     #

                   
                  Copyright: 1994 Petr Krysl

   Czech Technical University in Prague, Faculty of Civil Engineering,
      Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                  email: pk@power2.fsv.cvut.cz
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "Esimple.h"


#ifndef min
#    define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#    define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

typedef int (*EXVRML1_write_object_func) (GraphicObj *gp);

typedef struct EXVRML1_export_Methods {
  EXVRML1_write_object_func write;
}              EXVRML1_export_Methods;


/* Private data */
static int first = YES;
static EXVRML1_export_Methods *methods;
static FILE *in_stream;
static FILE *out_stream;
static double x, y, z;
#define DIM 200
static  WCRec pts[DIM];
static FPNum vals[DIM];
static  int np;
static char geomid[64];
static EGraphicId gtimeid = 0, gserid = 0, substid = 0;
static double R, G, B;


#define GET_PIXEL(GP, PIXEL) {  EGCopyAttributesToActiveSet (GP);       \
                                PIXEL = EASValsGetColor ();             \
                              }
#define GET_RGB(PIXEL, R, G, B)                         \
  {                                                     \
    ColorPixelToRGBStandard (PIXEL, &R, &G, &B);        \
  }


#define COORDS(ith) { x = pts[ith].x; y = pts[ith].y; z = pts[ith].z; }

#define OS out_stream

static int
boot_vrml_export (void);

#define SET_UP_OS(stream) { if (first) boot_vrml_export (); out_stream = stream; }


#define NL "\n"



typedef struct BBOX_t {
  WCRec ll, ur;
}              BBOX_t;

static int 
update_bbox (NODE ptr, NODE elem)
{
  WCRec ll, ur;
  BBOX_t *bbox = (BBOX_t *) ptr;

  EGGraphicsBoundingBoxWC ((GraphicObj *)elem, &ll, &ur);
  bbox->ll.x = min (ll.x, bbox->ll.x);
  bbox->ll.y = min (ll.y, bbox->ll.y);
  bbox->ll.z = min (ll.z, bbox->ll.z);
  bbox->ur.x = max (ur.x, bbox->ur.x);
  bbox->ur.y = max (ur.y, bbox->ur.y);
  bbox->ur.z = max (ur.z, bbox->ur.z);
  
  return 1;
}

static void
compute_bbox (EModel *model, BBOX_t  *bbox)
{
  bbox->ll.x = +FLT_MAX;
  bbox->ll.y = +FLT_MAX;
  bbox->ll.z = +FLT_MAX;
  bbox->ur.x = -FLT_MAX;
  bbox->ur.y = -FLT_MAX;
  bbox->ur.z = -FLT_MAX;
  EMDispatchToGraphicObjects (model, update_bbox, (caddrt)bbox);
}


static void
write_axes (FILE *stream, BBOX_t *bbox)
{
  double dim = (fabs (bbox->ur.x - bbox->ll.x) +
                fabs (bbox->ur.y - bbox->ll.y) +
                fabs (bbox->ur.z - bbox->ll.z)) / 10;
  
  fprintf (OS, "Separator {" NL);
  fprintf (OS, "  Info { string \"Axes\" }" NL);
  fprintf (OS, "  # rotate Z-axis up, X-axis out of screen" NL);
#if 0 /* this was supposed to rotate Z-axis up; well it didn't work.  may when I have some time... */
  fprintf (OS, "  Rotation { rotation -1 0 0 1.57079632679489661923 }" NL);
  fprintf (OS, "  Rotation { rotation 0 0 -1 1.57079632679489661923 }" NL);
#endif
  fprintf (OS, "  Translation { translation %g %g %g }" NL,
           (bbox->ur.x + bbox->ll.x) /2,
           (bbox->ur.y + bbox->ll.y) /2,
           (bbox->ur.z + bbox->ll.z) /2);
  fprintf (OS, "  DEF ScaleToPixels Separator {Scale { scaleFactor %g %g %g }}" NL, dim, dim, dim);
  fprintf (OS, "  # sphere at the center" NL);
  fprintf (OS, "  Separator {" NL);
  fprintf (OS, "    Material {" NL);
  fprintf (OS, "      diffuseColor 1 1 1" NL);
  fprintf (OS, "      emissiveColor 0.2 0.2 0.2" NL);
  fprintf (OS, "      specularColor  0.5 0.1 0.1" NL);
  fprintf (OS, "      transparency 0" NL);
  fprintf (OS, "    }" NL);
  fprintf (OS, "    Sphere { radius 0.04 }" NL);
  fprintf (OS, "  }" NL);
  fprintf (OS, "  # y-axis" NL);
  fprintf (OS, "  Separator {" NL);
  fprintf (OS, "    Material {" NL);
  fprintf (OS, "      diffuseColor 0 1 0" NL);
  fprintf (OS, "      emissiveColor 0 0.1 0 " NL);
  fprintf (OS, "      specularColor  0.1 0.5 0.1" NL);
  fprintf (OS, "      transparency 0.5" NL);
  fprintf (OS, "    }" NL);
  fprintf (OS, "    DEF AxisArrow Separator {" NL);
  fprintf (OS, "      Translation { translation 0 0.5 0 }" NL);
  fprintf (OS, "      Cylinder { radius 0.02 height  1 }" NL);
  fprintf (OS, "      Translation { translation 0 0.5 0 }" NL);
  fprintf (OS, "      Cone { bottomRadius 0.04 height 0.2 }" NL);
  fprintf (OS, "      Translation { translation 0 -0.5 0 }" NL);
  fprintf (OS, "      Translation { translation 0 -0.5 0 }" NL);
  fprintf (OS, "    }" NL);
  fprintf (OS, "  }" NL);
  fprintf (OS, "  # x-axis" NL);
  fprintf (OS, "  Separator {" NL);
  fprintf (OS, "    Material {" NL);
  fprintf (OS, "      diffuseColor 1 0 0" NL);
  fprintf (OS, "      emissiveColor 0.1 0 0" NL);
  fprintf (OS, "      specularColor  0.5 0.1 0.1" NL);
  fprintf (OS, "      transparency 0.5" NL);
  fprintf (OS, "    }" NL);
  fprintf (OS, "    Transform { translation 0 0 0" NL);
  fprintf (OS, "                rotation 0 0 -1  1.57079632679489661923 }" NL);
  fprintf (OS, "    USE AxisArrow" NL);
  fprintf (OS, "    Transform { translation 0 0 0" NL);
  fprintf (OS, "                rotation 0 0 -1 -1.57079632679489661923 }" NL);
  fprintf (OS, "  }" NL);
  fprintf (OS, " # z-axis" NL);
  fprintf (OS, "  Separator {" NL);
  fprintf (OS, "    Material {" NL);
  fprintf (OS, "      diffuseColor 0 0 1" NL);
  fprintf (OS, "      emissiveColor 0 0 0.1 " NL);
  fprintf (OS, "      specularColor  0.1 0.1 0.5" NL);
  fprintf (OS, "      transparency 0.5" NL);
  fprintf (OS, "    }" NL);
  fprintf (OS, "    Transform { translation 0 0 0" NL);
  fprintf (OS, "                rotation 1 0 0 1.57079632679489661923 }" NL);
  fprintf (OS, "    USE AxisArrow " NL);
  fprintf (OS, "    Transform { translation 0 0 0" NL);
  fprintf (OS, "                rotation 1 0 0 -1.57079632679489661923 }" NL);
  fprintf (OS, "  }" NL);
  fprintf (OS, "  Scale { scaleFactor %g %g %g }" NL, 1/dim, 1/dim, 1/dim);
  fprintf (OS, "  Translation { translation %g %g %g }" NL,
           -(bbox->ur.x + bbox->ll.x) /2,
           -(bbox->ur.y + bbox->ll.y) /2,
           -(bbox->ur.z + bbox->ll.z) /2);
  fprintf (OS, "}" NL);
}

static LIST objs = NULL;
static EModel *mdl = NULL;
static dict_dictionary mats = NULL;


static int
write_graphic_obj (FILE *stream,
                   EModel *model,
                   GraphicObj *gp)
{
  EGraphicType type;
  
  SET_UP_OS (stream);

  type       = EGGraphicType (gp);
  if (type < EG_MAX_METHOD_TABLE && methods[type].write) {
    EGCopyAttributesToActiveSet (gp);
    (*methods[type].write) (gp);
  }
  return TRUE;
}

static char matnamebuf[256];

static char *
mat_name (EPixel pixel)
{
  int i, len;
  sprintf (matnamebuf, "%s", ColorGetStringFromPixel (pixel));
  for (i = 0, len = strlen (matnamebuf); i < len; i++) {
    if (matnamebuf[i] == ' ') matnamebuf[i] = '_';
    if (matnamebuf[i] == '#') matnamebuf[i] = 'M';
  }
  return matnamebuf;
}

static int
write_material (EPixel pixel)
{
  GET_RGB (pixel, R, G, B);
  fprintf (OS, "DEF %s Material {" NL, mat_name (pixel));
  fprintf (OS, "   ambientColor %g %g %g" NL, R, G, B);
  fprintf (OS, "   diffuseColor %g %g %g" NL, R, G, B);
  fprintf (OS, "   specularColor  0.5 0.5 0.5" NL);
  fprintf (OS, "   transparency 0" NL);
  fprintf (OS, "}" NL);
  return TRUE;
}

int
EXVRMLClose (FILE *stream)
{
  GraphicObj *go;
  dict_iterator diter;
  void *p;
  
  diter = dict_get_iterator ();
  dict_reset_iterator (diter);
  while ((p = dict_iter_objs (mats, diter))) {
    write_material ((EPixel)p);
  }
  dict_del_iterator (diter);

  go = NULL;
  while ((go = (GraphicObj *) get_list_next (objs, go))) {
    write_graphic_obj (stream, mdl, go);
  }
  
  SET_UP_OS (stream);
  fprintf (OS, "}" NL);
  fprintf (OS, " # end-of-data" NL);

  dict_destroy_dictionary (mats);
  mats = NULL;
  
  return 1;
}


static int
shrink (WCRec *pts, int npts, double shrinkf)
{
  WCRec c = {0,0,0};
  int i;
  for (i = 0; i < npts; i++) {
    c.x += pts[i].x;
    c.y += pts[i].y;
    c.z += pts[i].z;
  }
  c.x /= npts; c.y /= npts; c.z /= npts;
  for (i = 0; i < npts; i++) {
    pts[i].x = shrinkf * pts[i].x + (1 - shrinkf) * c.x;
    pts[i].y = shrinkf * pts[i].y + (1 - shrinkf) * c.y;
    pts[i].z = shrinkf * pts[i].z + (1 - shrinkf) * c.z;
  }
  return TRUE;
}

static int
write_coords (GraphicObj *gp)
{
  int i;
  double shrinkf = EASValsGetShrink ();
  
  if (!EGGetGraphicsGeometry(gp, pts, &np))
    return FALSE;

  if (shrinkf != 1) shrink (pts, np, shrinkf);
  
  fprintf (OS, "Coordinate3 { point [" NL);
  for (i = 0; i < np; i++) {
    fprintf (OS, "%g %g %g,\n", pts[i].x, pts[i].y, pts[i].z);
  }
  fprintf (OS, "] }" NL);
  return TRUE;
}


int
EXVRMLWriteGeom (FILE *stream,
                EModel *model,
                GraphicObj *gp)
{
  EPixel pixel;
  GET_PIXEL (gp, pixel);
  if (!dict_find_obj_with_handle_name (mats, (void *)pixel))
    dict_add_obj_handle_name (mats, (void *)pixel, (void *)pixel);
  add_to_tail (objs, gp);
  return TRUE;
}







static int 
write_quad (GraphicObj *gp)
{
  EPixel pixel;
  GET_PIXEL (gp, pixel);

  fprintf (OS, "USE %s" NL, mat_name (pixel));
  
  if (!write_coords (gp))
    return FALSE;

  fprintf (OS, "IndexedFaceSet{coordIndex[0,1,2,3]}" NL);

  return TRUE;
}


static int 
write_line (GraphicObj *gp)
{
  EPixel pixel;
  GET_PIXEL (gp, pixel);

  fprintf (OS, "USE %s" NL, mat_name (pixel));
  
  if (!write_coords (gp))
    return FALSE;

  fprintf (OS, "IndexedLineSet{coordIndex[0,1]}" NL);

  return TRUE;
}





static int 
write_tri (GraphicObj *gp)
{
  EPixel pixel;
  GET_PIXEL (gp, pixel);

  fprintf (OS, "USE %s" NL, mat_name (pixel));
  
  if (!write_coords (gp))
    return FALSE;

  fprintf (OS, "IndexedFaceSet{coordIndex[0,1,2]}" NL);
  
  return TRUE;
}




static int 
write_tet (GraphicObj *gp)
{
  EPixel pixel;
  GET_PIXEL (gp, pixel);

  fprintf (OS, "USE %s" NL, mat_name (pixel));
  
  if (!write_coords (gp))
    return FALSE;

  fprintf (OS, "IndexedFaceSet{coordIndex[0,2,1,-1,0,1,3,-1,1,2,3,-1,2,0,3]}" NL);

  return YES;
}



static int 
write_hex (GraphicObj *gp)
{
  EPixel pixel;
  GET_PIXEL (gp, pixel);

  fprintf (OS, "USE %s" NL, mat_name (pixel));
  
  if (!write_coords (gp))
    return FALSE;

  fprintf (OS, "IndexedFaceSet{coordIndex[0,3,2,1,-1,0,1,5,4,-1,1,2,6,5,-1,2,3,7,6,-1,3,0,4,7,-1,4,5,6,7]}" NL);

  return YES;
}


static int 
write_tet10 (GraphicObj *gp)
{
  EPixel pixel;
  GET_PIXEL (gp, pixel);

  fprintf (OS, "USE %s" NL, mat_name (pixel));
  
  if (!write_coords (gp))
    return FALSE;

  fprintf (OS, "IndexedFaceSet{coordIndex[0,6,4,-1,6,2,5,-1,5,1,4,-1,6,5,4,-1,0,4,7,-1,4,1,8,-1,8,3,7,-1,4,8,7,-1,1,5,8,-1,5,2,9,-1,9,3,8,-1,5,9,8,-1,2,6,9,-1,6,0,7,-1,7,3,9,-1,6,7,9]}" NL);

  return YES;
}



static int 
write_tri6 (GraphicObj *gp)
{
  EPixel pixel;
  GET_PIXEL (gp, pixel);

  fprintf (OS, "USE %s" NL, mat_name (pixel));
  
  if (!write_coords (gp))
    return FALSE;

  fprintf (OS, "IndexedFaceSet{coordIndex[0,3,5,-1,3,1,4,-1,4,2,5,-1,3,4,5]}" NL);

  return YES;
}

static int 
write_marker (GraphicObj *gp)
{
  EPixel pixel;
  WCRec pts[1];
  int np;

  if (!EGGetGraphicsGeometry(gp, pts, &np))
    return FALSE;

  GET_PIXEL (gp, pixel);

  fprintf (OS, "Separator{USE %s\nTranslation{translation %f %f %f}\nUSE ScaleToPixels\nSphere{radius 0.02}}" NL,
           mat_name(pixel), pts[0].x, pts[0].y, pts[0].z);

  return YES;
}


static int
boot_vrml_export (void)
{
#define SET_ENTRY(which, WRITE)           \
  { methods[which].write = WRITE;         \
  }

  methods = (EXVRML1_export_Methods *) make_node (sizeof(EXVRML1_export_Methods) * EG_MAX_METHOD_TABLE);
  if (!methods)
    EUFailedMakeNode("boot_vrml_export");

  SET_ENTRY(EG_UNKNOWN_TYPE, NULL);
  SET_ENTRY(EG_LINE3D, write_line);
  SET_ENTRY(EG_RECTANGLE_VC, NULL);
  SET_ENTRY(EG_MARKER3D, write_marker);
  SET_ENTRY(EG_ANNTEXT3D, NULL);
  SET_ENTRY(EG_GGROUP, NULL);
  SET_ENTRY(EG_CTLBLOCK, NULL);
  SET_ENTRY(EG_TRIANGLE3D, write_tri);
  SET_ENTRY(EG_CBEZCURVE3D, NULL);
  SET_ENTRY(EG_BCBEZSURFACE3D, NULL);
  SET_ENTRY(EG_TRIANGLEWD3D, NULL);
  SET_ENTRY(EG_QUADWD3D, NULL);
  SET_ENTRY(EG_TETRAHEDRON, write_tet);
  SET_ENTRY(EG_RIGHT_ANGLE_PRISM, NULL);
  SET_ENTRY(EG_BCB_HEXAHEDRON, NULL);
  SET_ENTRY(EG_QUAD3D, write_quad);
  SET_ENTRY(EG_PYRAMID, NULL);
  SET_ENTRY(EG_HEXAHEDRON, write_hex);
  SET_ENTRY(EG_VECMARKER3D, NULL);
  SET_ENTRY(EG_VECTOR3D, NULL);
  SET_ENTRY(EG_RBEZCURVE3D, NULL);
  SET_ENTRY(EG_RBEZSURFACE3D, NULL);
  SET_ENTRY(EG_HEXAHEDRONWD, NULL);
  SET_ENTRY(EG_TETRAHEDRONWD, NULL);
  SET_ENTRY(EG_TET10, write_tet10);
  SET_ENTRY(EG_TRI6, write_tri6);
  first = NO;
  return 1;
}


int
EXVRMLStart (EModel *model, FILE *stream)
{
  BBOX_t bbox;

  mdl = model;

  if (!mats) mats = dict_create_dictionary ();

  compute_bbox (model, &bbox);
  
  if (!objs) objs = make_list ();
  flush_list (objs, NOT_NODES);
  
  SET_UP_OS (stream);
  fprintf (OS, "#VRML V1.0 ascii" NL);
  fprintf (OS, "# " NL);
  fprintf (OS, "# Elixir (C) 1994-98 Petr Krysl" NL);
  fprintf (OS, "ShapeHints {\n"
               "   vertexOrdering  CLOCKWISE\n"
               "   shapeType       UNKNOWN_SHAPE_TYPE\n"
               "   faceType        CONVEX\n"
               "   creaseAngle     0.5\n"
               "}" NL);
  fprintf (OS, "Separator {" NL);
  fprintf (OS, " # start-of-data" NL);
  write_axes (stream, &bbox);
  return 1;
}
