/*
 
              ####### #         ###   #     #   ###   ######
              #       #          #     #   #     #    #     #
              #       #          #      # #      #    #     #
              #####   #          #       #       #    ######
              #       #          #      # #      #    #   #
              #       #          #     #   #     #    #    #
              ####### #######   ###   #     #   ###   #     #
 
                   
                  Copyright: 1994 Petr Krysl
 
   Czech Technical University in Prague, Faculty of Civil Engineering,
      Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                  email: pk@power2.fsv.cvut.cz
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
*/

#include "Esimple.h"

#define HEXSE_STATUS_MESSAGE 1000
#define HEXSE_ERROR_MESSAGE  2000


#ifndef min
#    define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#    define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

void
make_from_geom (char *name);

void
make_from_dxf (char *name);

void
make_from_fct (char *name);

void
make_from_gts (char *name);



static int
has_suffix (char *f, char *s /* suffix */)
{
        int lf, ls, i;

        if (!f || !s)
                return FALSE;

        lf = strlen((f));
        ls = min(strlen((s)), lf);
        for (i = 0; i < ls; i++)
                if (toupper((f)[lf-i]) != toupper((s)[ls-i]))
                        return FALSE;
        return TRUE;
}

#define MAXCOLR 11
static int colr = 0;
static char *color[MAXCOLR] = {
                                      "color cyan",
                                      "color yellow",
                                      "color red",
                                      "color green",
                                      "color orange",
                                      "color slateblue",
                                      "color hotpink",
                                      "color blueviolet",
                                      "color greenyellow"
                              };

static char *
filenameminussuffix(char *f)
{
        int maxfnl=2048;
        static char fnms[2048];
        int i, sp=0;
        for (i=strlen(f); i>=0; i--) {
                if (f[i] == '.') {
                        sp=i;
                        break;
                }
        }
        if (sp <= 0)
                sp=strlen(f);
        for (i=0; i<min(maxfnl, sp); i++)
                fnms[i]=f[i];
        return fnms;
}

static int didcmdfile=0;

static void
exec_cmd_file (char *name) 
{
  const char *linebuf;
  FILE *stream;
  ckit_iobuf_t IOB = NULL;

  didcmdfile=1;

  stream = ckit_io_open_stream (name, "r");
  if (stream) {
    IOB    = ckit_io_new_buffer (stream);
    while ((linebuf = ckit_io_scan_line (IOB, FALSE)) ) {
      ESIHandleCmd ((char *) linebuf);
    }  
    ckit_io_delete_buffer (IOB);
    fclose (stream);
  }   else   {
    fprintf (stderr, "\nCould not open command file \"%s\\n", name);
    exit (1);
  }  
}  

static void
suck_in_file (char *f)
{
        char command[1024];

        ESIHandleCmd (color[colr++]);
        if (colr == MAXCOLR)
                colr = 0;

        if        (has_suffix (f, ".dxf")) {
                make_from_dxf (f);
        } else if (has_suffix (f, ".geom")) {
                make_from_geom (f);
        } else if (has_suffix (f, ".gts")) {
                make_from_gts (f);
        } else if (has_suffix (f, ".fct")) {
                make_from_fct (f);
        } else if (has_suffix (f, ".cmd")) {
                exec_cmd_file (f);
        } else if (has_suffix (f, ".egf")) {
                sprintf (command, "LOAD \"%s\"", f);
                ESIHandleCmd (command);
        } else
                fprintf (stderr, "Did not recognize format of \"%s\"\n", f);
}

int
main(int argc, char **argv)
{
        int i;
        int saveegf = 0;
        char egffile[2048];
        unsigned long m = 0;
        m |= ESI_GRAPHIC_EDITOR_MASK;
        ESIBuildInterface(m, argc, argv);
        ESIPopup();
        i = 1;
        while (i < argc) {
                if (ckit_io_file_exists (argv[i])) {
                        suck_in_file (argv[i]);
                        sprintf(egffile, "%s.egf", filenameminussuffix(argv[i]));
                }
                if (strcmp(argv[i],"-saveegf") == 0)
                        saveegf=1;
                i++;
        }
        if (saveegf) {
                OnDiskSetFileName(egffile);
                EMSaveAllGraphics(ESIModel());
		return 1;
        }
        if (!didcmdfile) ESIHandleCmd("COLOR yellow");
        if (!didcmdfile) ESIHandleCmd("SHRINK 0.9");
        if (!didcmdfile) ESIHandleCmd("NEWFRAME width 800 height 800");
        if (!didcmdfile) ESIHandleCmd("view bground white");
        ESIPopupAndRun();
        return 1;
}







static void
read_cmd_CB (Widget w, XtPointer ptr, XtPointer call)
{
        char *name;
        FILE *fp;

        fp = ESISelFile("Select Elixir command file", "OK", "CANCEL",
                        NULL, NULL, "r", &name);
        if (fp != NULL) {
                fclose(fp);
               exec_cmd_file (name);
        }
}

static void
make_from_geom_CB (Widget w, XtPointer ptr, XtPointer call)
{
        char *name;
        FILE *fp;

        fp = ESISelFile("Select GEOM file", "OK", "CANCEL",
                        NULL, NULL, "r", &name);
        if (fp != NULL) {
                fclose(fp);
                make_from_geom (name);
        }
}

static void
make_from_gts_CB (Widget w, XtPointer ptr, XtPointer call)
{
        char *name;
        FILE *fp;

        fp = ESISelFile("Select GTS file", "OK", "CANCEL",
                        NULL, NULL, "r", &name);
        if (fp != NULL) {
                fclose(fp);
                make_from_gts (name);
        }
}

static void
make_from_dxf_CB (Widget w, XtPointer ptr, XtPointer call)
{
        char *name;
        FILE *fp;

        fp = ESISelFile("Select DXF file", "OK", "CANCEL",
                        NULL, NULL, "r", &name);
        if (fp != NULL) {
                fclose(fp);
                make_from_dxf (name);
        }
}

static void
make_from_fct_CB (Widget w, XtPointer ptr, XtPointer call)
{
        char *name;
        FILE *fp;

        fp = ESISelFile("Select facets (FCT) file", "OK", "CANCEL",
                        NULL, NULL, "r", &name);
        if (fp != NULL) {
                fclose(fp);
                make_from_fct (name);
        }
}


static void
destroy_all_CB(Widget w, XtPointer ptr, XtPointer call)
{
        EMDestroyAllGraphics(ESIModel());
}


void
ESICustomize(Widget parent_pane)
{
        Arg args[3];
        int argn;
        Widget b;

        argn = 0;

        b    = ESIAddButton("read_geom_file", "READ \"CMD\" FILE ...",
                            commandWidgetClass,
                            parent_pane, args, argn, read_cmd_CB, NULL);
        b    = ESIAddButton("read_geom_file", "READ \"GEOM\" FILE ...",
                            commandWidgetClass,
                            parent_pane, args, argn, make_from_geom_CB, NULL);
        b    = ESIAddButton("read_gts_file", "READ \"GTS\" FILE ...",
                            commandWidgetClass,
                            parent_pane, args, argn, make_from_gts_CB, NULL);
        b    = ESIAddButton("read_dxf_file", "READ \"DXF\" FILE ...",
                            commandWidgetClass,
                            parent_pane, args, argn, make_from_dxf_CB, NULL);
        b    = ESIAddButton("read_fct_file", "READ \"FCT\" FILE ...",
                            commandWidgetClass,
                            parent_pane, args, argn, make_from_fct_CB, NULL);
        b    = ESIAddButton("destroy_curr_model", "DESTROY ALL GRAPHICS",
                            commandWidgetClass,
                            parent_pane, args, argn, destroy_all_CB, NULL);
}
