
#include "Eprimtv.h"

#ifndef min
#    define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#    define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

static int 
next_point_wd(FPNum sx, FPNum sy, FPNum sz,
              FPNum ex, FPNum ey, FPNum ez,
              FPNum sv, FPNum ev, FPNum lower, FPNum delta,
              FPNum *nx, FPNum *ny, FPNum *nz, FPNum *ns, FPNum *ne);
static int 
fill_triangle_with_colors(EView *v_p,
			  Display *dsp, Window win, GC agc,
			  FPNum *x, FPNum *y, FPNum *z, FPNum *values,
			  EBoxVC2DRec *view_box);
static void 
indices_seq(FPNum *values, int *indx1, int *indx2, int *indx3);
static int 
compute_poly_bbox(FPNum *x, FPNum *y, int npoints, EBoxVC2DRec *box);

void 
EVTriangleWD3DPrimitive (EView        *v_p,
                         GraphicObj   *g_p, 
                         WCRec        *f,
                         WCRec        *s,
                         WCRec        *t,
                         double        val1,
                         double        val2,
                         double        val3,
                         EPixel        pixel,
                         int           fill,
                         int           edge_flag,
                         EPixel        edge_pixel,
                         float         shrink,
                         EDrawMode     draw_mode)
{
  int x1, x2, y1, y2, x3, y3, x4, y4;
  VCRec vc[4], ll, ur, tll, tur; 
  EBoxVC2DRec vbox, tbox;
  WCRec sf, ss, st;
  double tmp;
  EV_line_end_point_rec p1, p2, p3;
  EV_triangle_clip_result clip_result = EV_TRIANGLE_IS_IN;
  EV_line_end_point_rec  pnts_return[4];
  int npnts = 3;
  double data[4];
  VCRec model_clip_plane_center, model_clip_plane_normal;

  EV_display                    = XtDisplay(v_p->view_widget);
  EV_curr_win                   = v_p->draw_into;
  EV_should_clip_by_view_planes = v_p->should_clip_by_fg_bg_view_planes;
  EV_fg_plane_depth             = v_p->fg_view_plane_n;
  EV_bg_plane_depth             = v_p->bg_view_plane_n;
    
  if (shrink != 1) {
    COPY_WCREC (&sf, f);
    COPY_WCREC (&ss, s);
    COPY_WCREC (&st, t);
    SHRINK_TRIANGLE_DATA (shrink, sf.x, ss.x, st.x);
    SHRINK_TRIANGLE_DATA (shrink, sf.y, ss.y, st.y);
    SHRINK_TRIANGLE_DATA (shrink, sf.z, ss.z, st.z);
    SHRINK_TRIANGLE_DATA (shrink, val1, val2, val3);
    EVWCtoVC (v_p, &sf, &vc[0]);
    EVWCtoVC (v_p, &ss, &vc[1]);
    EVWCtoVC (v_p, &st, &vc[2]);
  } else {
    EVWCtoVC (v_p, f, &vc[0]);
    EVWCtoVC (v_p, s, &vc[1]);
    EVWCtoVC (v_p, t, &vc[2]);
  }
  data[0] = val1;
  data[1] = val2;
  data[2] = val3;
  
  if (v_p->should_clip_by_model_planes  && EGIsClippable (g_p)) {
    if (v_p->model_clip_plane.clip_is_on) {
      WCRec Zero = {0, 0, 0};
      TransfPntWCToVC (&(v_p->VectorU), &(v_p->VectorV), &(v_p->Normal),
                       &(v_p->VCOrigin), &v_p->model_clip_plane.center,
                       &model_clip_plane_center);
      TransfPntWCToVC (&(v_p->VectorU), &(v_p->VectorV), &(v_p->Normal),
                       &Zero, &v_p->model_clip_plane.normal,
                       &model_clip_plane_normal);
      COPY_VCREC (&p1.loc.vc, &vc[0]); p1.data = data[0];
      COPY_VCREC (&p2.loc.vc, &vc[1]); p2.data = data[1];
      COPY_VCREC (&p3.loc.vc, &vc[2]); p3.data = data[2];
      clip_result
        = clip_triangle_by_a_plane (TRUE, &p1, &p2, &p3,
                                    (WCRec *)&model_clip_plane_center,
                                    (WCRec *)&model_clip_plane_normal,
                                    pnts_return, &npnts);
      if (clip_result == EV_TRIANGLE_IS_OUT)
        return;
      if (clip_result == EV_TRIANGLE_CLIPPED) {
        if (EGInvisibleWhenClipped (g_p))
          return;
        COPY_VCREC (&vc[0], &pnts_return[0].loc.vc);
        data[0] = pnts_return[0].data;
        COPY_VCREC (&vc[1], &pnts_return[1].loc.vc);
        data[1] = pnts_return[1].data;
        COPY_VCREC (&vc[2], &pnts_return[2].loc.vc);
        data[2] = pnts_return[2].data;
        if (npnts == 4) {
          COPY_VCREC (&vc[3], &pnts_return[3].loc.vc);
          data[3] = pnts_return[3].data;
        }
      }
    }
  }
  
  EVGetViewBox (v_p, &ll, &ur);
  tmp = min(vc[0].u, vc[1].u); tll.u = min(tmp, vc[2].u);
  tmp = min(vc[0].v, vc[1].v); tll.v = min(tmp, vc[2].v);
  tmp = min(vc[0].n, vc[1].n); tll.n = min(tmp, vc[2].n);
  tmp = max(vc[0].u, vc[1].u); tur.u = max(tmp, vc[2].u);
  tmp = max(vc[0].v, vc[1].v); tur.v = max(tmp, vc[2].v);
  tmp = max(vc[0].n, vc[1].n); tur.n = max(tmp, vc[2].n);
  vbox.left  = ll.u; vbox.bottom  = ll.v;
  vbox.right = ur.u; vbox.top     = ur.v;
  tbox.left  = tll.u; tbox.bottom = tll.v;
  tbox.right = tur.u; tbox.top    = tur.v;
  
  if (BoxesOverlap (&tbox, &vbox)) {
    EV_render_mode = v_p->render_mode;
    EV_shade_mode  = v_p->shade_mode;
    if      (draw_mode == DRAW_MODE_DRAW) {
      EV_curr_GC    = v_p->writableGC;
      XSetForeground(EV_display, EV_curr_GC, pixel);
      XSetLineAttributes(EV_display, EV_curr_GC,
			 (unsigned)0, LineSolid, CapButt, JoinMiter);
    } else if (draw_mode == DRAW_MODE_ERASE) {
      EV_curr_GC    = v_p->eraseGC;
      EV_curr_pixel = v_p->background;
      XSetLineAttributes(EV_display, EV_curr_GC,
			 (unsigned)0, LineSolid, CapButt, JoinMiter);
    } else if (draw_mode == DRAW_MODE_HILITE) {
      EV_curr_GC    = v_p->hiliteGC;
      EV_curr_pixel = v_p->foreground;
    } else if (draw_mode == DRAW_MODE_UNHILITE) {
      EV_curr_GC    = v_p->erasehiliteGC;
      EV_curr_pixel = v_p->background;
    } else if (draw_mode == DRAW_MODE_XORDRAW) {
      EV_curr_GC    = v_p->defaultXORGC;
      EV_curr_pixel = v_p->foreground;
    }
    
    if ((EV_render_mode == FILLED_HIDDEN_RENDERING ||
	 EV_render_mode == CONST_SHADING_RENDERING) && fill != FILL_HOLLOW) {
      if (EV_render_mode == CONST_SHADING_RENDERING) {
	VCRec normal, v1, v2;
	v1.u = vc[1].u - vc[0].u;
        v1.v = vc[1].v - vc[0].v;
        v1.n = vc[1].n - vc[0].n;
	v2.u = vc[2].u - vc[0].u;
        v2.v = vc[2].v - vc[0].v;
        v2.n = vc[2].n - vc[0].n;
	CrossProd3((WCRec *)&normal, (WCRec *)&v1, (WCRec *)&v2);
	if (NormalizeVect3((WCRec *)&normal)) {
	  if (normal.n < 0) {
	    normal.u = -normal.u;
	    normal.v = -normal.v;
	    normal.n = -normal.n;
	  }
	  EV_curr_cosine = EVGetLightDirVsNormalCos (v_p, &normal);
        } 
      }  
      
      EVVCtoDC (v_p, &vc[0], &x1, &y1);
      EVVCtoDC (v_p, &vc[1], &x2, &y2);
      EVVCtoDC (v_p, &vc[2], &x3, &y3);
      TFCI (x1, y1, vc[0].n, x2, y2, vc[1].n, x3, y3, vc[2].n,
            data[0], data[1], data[2], v_p->fringe_table,
            v_p->view_dims_DC.x, v_p->view_dims_DC.y);
      if (clip_result == EV_TRIANGLE_CLIPPED && npnts == 4) {
        EVVCtoDC (v_p, &vc[3], &x4, &y4);
        TFCI (x1, y1, vc[0].n, x3, y3, vc[2].n, x4, y4, vc[3].n, 
              data[0], data[2], data[3], v_p->fringe_table,
              v_p->view_dims_DC.x, v_p->view_dims_DC.y);
      }
      if (edge_flag) {
        double n1, n2, n3, n4, dn;
        EV_curr_pixel = edge_pixel;
        dn = DELTA_N();
        n1 = vc[0].n + dn;
        n2 = vc[1].n + dn;
        n3 = vc[2].n + dn;
        if (clip_result == EV_TRIANGLE_IS_IN) {
          BHM(x1, y1, n1, x2, y2, n2);
          BHM(x1, y1, n1, x3, y3, n3);
          BHM(x2, y2, n2, x3, y3, n3);
        } else if (clip_result == EV_TRIANGLE_CLIPPED && npnts == 3) {
          BHM(x1, y1, n1, x2, y2, n2);
          BHM(x1, y1, n1, x3, y3, n3);
        } else if (clip_result == EV_TRIANGLE_CLIPPED && npnts == 4) {
          n4 = vc[3].n + dn;
          BHM(x1, y1, n1, x2, y2, n2);
          BHM(x3, y3, n3, x4, y4, n4);
          BHM(x1, y1, n1, x4, y4, n4);
        }
      }
    } else { /* Not hidden surface/line rendering */
      XPoint points[4];
      int ix, iy, np=3;
      
      if (EV_render_mode == WIRE_RENDERING || fill == FILL_HOLLOW ||
	  draw_mode == DRAW_MODE_XORDRAW) {
	int ix1, ix2, iy1, iy2, ix3, iy3;
	VCRec F, S;
	int draw_line_instead = NO;
        
#define PNTS_IDENTCL(_x1, _y1, _x2, _y2) (_x1==_x2 && _y1==_y2)
        
	if (draw_mode == DRAW_MODE_XORDRAW) {
	  EVVCtoDC(v_p, &vc[0], &ix1, &iy1);
	  EVVCtoDC(v_p, &vc[1], &ix2, &iy2);
	  EVVCtoDC(v_p, &vc[2], &ix3, &iy3);
	  if (PNTS_IDENTCL(ix1, iy1, ix2, iy2)) {
	    XDrawLine(EV_display, EV_curr_win, EV_curr_GC, ix1, iy1, ix3, iy3);
	    draw_line_instead = YES;
	  } else if (PNTS_IDENTCL(ix3, iy3, ix2, iy2)) {
	    XDrawLine(EV_display, EV_curr_win, EV_curr_GC, ix3, iy3, ix1, iy1);
	    draw_line_instead = YES;
	  } else if (PNTS_IDENTCL(ix1, iy1, ix3, iy3)) {
	    XDrawLine(EV_display, EV_curr_win, EV_curr_GC, ix2, iy2, ix3, iy3);
	    draw_line_instead = YES;
	  }
	}
        if (!draw_line_instead) {
	  F.u = vc[0].u; F.v = vc[0].v; F.n = vc[0].n;
	  S.u = vc[1].u; S.v = vc[1].v; S.n = vc[1].n; 
	  if (ClipLineToView(ll.u, ll.v, ur.u, ur.v,
			     &F.u, &F.v, &F.n, &S.u, &S.v, &S.n)) {
	    EVVCtoDC(v_p, &F, &ix1, &iy1);
	    EVVCtoDC(v_p, &S, &ix2, &iy2);
	    XDrawLine(EV_display, EV_curr_win, EV_curr_GC, ix1, iy1, ix2, iy2);
	  }
	  F.u = vc[1].u; F.v = vc[1].v; F.n = vc[1].n;
	  S.u = vc[2].u; S.v = vc[2].v; S.n = vc[2].n; 
	  if (ClipLineToView(ll.u, ll.v, ur.u, ur.v,
			     &F.u, &F.v, &F.n, &S.u, &S.v, &S.n)) {
	    EVVCtoDC(v_p, &F, &ix1, &iy1);
	    EVVCtoDC(v_p, &S, &ix2, &iy2);
	    XDrawLine(EV_display, EV_curr_win, EV_curr_GC, ix1, iy1, ix2, iy2);
	  }
          if (   (clip_result == EV_TRIANGLE_IS_IN)
              || (clip_result == EV_TRIANGLE_CLIPPED && npnts == 3)) {
            F.u = vc[0].u; F.v = vc[0].v; F.n = vc[0].n;
            S.u = vc[2].u; S.v = vc[2].v; S.n = vc[2].n; 
            if (ClipLineToView(ll.u, ll.v, ur.u, ur.v,
                               &F.u, &F.v, &F.n, &S.u, &S.v, &S.n)) {
              EVVCtoDC(v_p, &F, &ix1, &iy1);
              EVVCtoDC(v_p, &S, &ix2, &iy2);
              XDrawLine(EV_display, EV_curr_win, EV_curr_GC,
                        ix1, iy1, ix2, iy2);
            }
          } else if (clip_result == EV_TRIANGLE_CLIPPED && npnts == 4) {
            F.u = vc[3].u; F.v = vc[3].v; F.n = vc[3].n;
            S.u = vc[2].u; S.v = vc[2].v; S.n = vc[2].n; 
            if (ClipLineToView(ll.u, ll.v, ur.u, ur.v,
                               &F.u, &F.v, &F.n, &S.u, &S.v, &S.n)) {
              EVVCtoDC(v_p, &F, &ix1, &iy1);
              EVVCtoDC(v_p, &S, &ix2, &iy2);
              XDrawLine(EV_display, EV_curr_win, EV_curr_GC,
                        ix1, iy1, ix2, iy2);
            }
            F.u = vc[0].u; F.v = vc[0].v; F.n = vc[0].n;
            S.u = vc[3].u; S.v = vc[3].v; S.n = vc[3].n; 
            if (ClipLineToView(ll.u, ll.v, ur.u, ur.v,
                               &F.u, &F.v, &F.n, &S.u, &S.v, &S.n)) {
              EVVCtoDC(v_p, &F, &ix1, &iy1);
              EVVCtoDC(v_p, &S, &ix2, &iy2);
              XDrawLine(EV_display, EV_curr_win, EV_curr_GC,
                        ix1, iy1, ix2, iy2);
            }
	  }
	}
      } else { /* fill by plain Xlib */
        if (1) {
          FPNum x[3], y[3], z[3], v[3];
          
          x[0] = vc[0].u; y[0] = vc[0].v; z[0] = vc[0].n; v[0] = data[0];
          x[1] = vc[1].u; y[1] = vc[1].v; z[1] = vc[1].n; v[1] = data[1];
          x[2] = vc[2].u; y[2] = vc[2].v; z[2] = vc[2].n; v[2] = data[2];
          fill_triangle_with_colors (v_p, EV_display, EV_curr_win, EV_curr_GC,
                                     x, y, z, v, &vbox);
          if (clip_result == EV_TRIANGLE_CLIPPED && npnts == 4) {
            x[0] = vc[0].u; y[0] = vc[0].v; z[0] = vc[0].n; v[0] = data[0];
            x[1] = vc[2].u; y[1] = vc[2].v; z[1] = vc[2].n; v[1] = data[2];
            x[2] = vc[3].u; y[2] = vc[3].v; z[2] = vc[3].n; v[2] = data[3];
            fill_triangle_with_colors (v_p, EV_display, EV_curr_win,
                                       EV_curr_GC,
                                       x, y, z, v, &vbox);
          }
        } else {
          np = 3;
          EVVCtoDC(v_p, &vc[0], &ix, &iy);
          points[0].x = ix; points[0].y = iy;
          EVVCtoDC(v_p, &vc[1], &ix, &iy);
          points[1].x = ix; points[1].y = iy;
          EVVCtoDC(v_p, &vc[2], &ix, &iy);
          points[2].x = ix; points[2].y = iy;
          if (clip_result == EV_TRIANGLE_CLIPPED && npnts == 4) {
            np = 4;
            EVVCtoDC(v_p, &vc[3], &ix, &iy);
            points[3].x = ix; points[3].y = iy;
          }
          XFillPolygon(EV_display, EV_curr_win, EV_curr_GC,
                       points, np, Convex, CoordModeOrigin);
        }
      }
    }
  } /* if (BoxesOverlap(... */
}



static int 
fill_triangle_with_colors (EView *v_p,
                           Display *dsp, Window win, GC agc,
                           FPNum *x, FPNum *y, FPNum *z, FPNum *values,
                           EBoxVC2DRec *view_box)
{
  FPNum lisx, liisx, lisy, liisy, lisz,
  liisz, liex, liiex, liey, liiey, liez, liiez;
  FPNum svi, evi, svii, evii;
  FPNum nx, ny, nz, nsv, nev;
  int npoints;
  FPNum lower;
  FPNum value1, value2, value3;
  FPNum delta, val;
  int range;
  EPixel color;
  int indx1, indx2, indx3;
  FPNum tmpx[7], tmpy[7], tmpz[7];
  FPNum *pvx, *pvy, *pvz;
  FPNum eps;
#define EPSILON 0.00001
  Widget w;
  EBoxVC2DRec pbox;
  EFringeTable ft;
  
  w   = EVViewPToWidget(v_p);
  ft  = EVGetAssocFringeTable(v_p);
  
  indices_seq(values, &indx1, &indx2, &indx3);
  lisx   = x[indx1]; lisy   = y[indx1]; lisz   = z[indx1];
  liex   = x[indx2]; liey   = y[indx2]; liez   = z[indx2];
  liisx  = x[indx1]; liisy  = y[indx1]; liisz  = z[indx1];
  liiex  = x[indx3]; liiey  = y[indx3]; liiez  = z[indx3];
  value1 = values[indx1]; value2 = values[indx2]; value3 = values[indx3];
  eps    = EPSILON * (fabs(liex - lisx) + fabs(liiex - liisx) +
		      fabs(liey - lisy) + fabs(liiey - liisy) +
		      fabs(liez - lisz) + fabs(liiez - liisz)); 

  delta = ColorFringesDelta(ft);
  val   = value1;
  range = ColorFringeValueToRange(ft, val);
  color = ColorFringeRangeToColor(range);
  lower = ColorFringeClosestLowerValue(ft, val);
  svi   = value1; evi   = value2;
  svii  = value1; evii  = value3;
  pvx = tmpx; pvx--; pvy = tmpy; pvy--; pvz = tmpz; pvz--;
 continue_1:
  npoints = 0;
  next_point_wd(lisx, lisy, lisz, liex, liey, liez,
	      svi, evi, lower, delta, &nx, &ny, &nz, &nsv, &nev);
  if (PWDAPPOXEQ(liex, liey, liez, evi, nx, ny, nz, nev, eps)) {
    liex = liiex; liey = liiey; liez = liiez; evi  = evii;
    npoints = 2; 
    pvx[npoints] = nx; pvy[npoints] = ny; pvz[npoints] = nz;
    npoints = 3; 
    pvx[npoints] = lisx; pvy[npoints] = lisy; pvz[npoints] = lisz;
    lisx = nx; lisy = ny; lisz = nz; svi  = nsv;
    next_point_wd(lisx, lisy, lisz, liex, liey, liez,
		svi, evi, lower, delta, &nx, &ny, &nz, &nsv, &nev);
    npoints = 1;
    pvx[npoints] = nx; pvy[npoints] = ny; pvz[npoints] = nz;
    npoints = 3;
  } else {
    npoints = 1;
    pvx[npoints] = nx; pvy[npoints] = ny; pvz[npoints] = nz;
    npoints = 2;
    pvx[npoints] = lisx; pvy[npoints] = lisy; pvz[npoints] = lisz;
    npoints = 2;
  }
  lisx = nx; lisy = ny; lisz = nz; svi  = nsv; evi  = nev;
  npoints = npoints + 1;
  pvx[npoints] = liisx; pvy[npoints] = liisy; pvz[npoints] = liisz;
  next_point_wd(liisx, liisy, liisz, liiex, liiey, liiez,
	      svii, evii, lower, delta, &nx, &ny, &nz, &nsv, &nev);
  npoints = npoints + 1;
  pvx[npoints] = nx; pvy[npoints] = ny; pvz[npoints] = nz;
  if (PWDAPPOXEQ(liiex, liiey, liiez, evii, nx, ny, nz, nev, eps)) {
    liiex = liex; liiey = liey; liiez = liez; evii = evi;
    liisx = nx; liisy = ny; liisz = nz; svii = nsv;
    next_point_wd(liisx, liisy, liisz, liiex, liiey, liiez,
		svii, evii, lower, delta, &nx, &ny, &nz, &nsv, &nev);
    npoints = npoints + 1;
    pvx[npoints] = nx; pvy[npoints] = ny; pvz[npoints] = nz;
  }
  liisx = nx; liisy = ny; liisz = nz; svii  = nsv; evii  = nev;
  npoints = npoints + 1;
  pvx[npoints] = pvx[1]; pvy[npoints] = pvy[1]; pvz[npoints] = pvz[1];

  compute_poly_bbox(tmpx, tmpy, npoints, &pbox);
  if (BoxesOverlap(&pbox, view_box)) {
    XPoint points[7];
    int i;
    int ix, iy;
    VCRec ap;
    
    XSetForeground(dsp, agc, color);
    for (i = 0; i < npoints; i++) {
      ap.u = tmpx[i]; ap.v = tmpy[i]; ap.n = tmpz[i];
      EVVCtoDC(v_p, &ap, &(ix), &(iy));
      points[i].x = ix; points[i].y = iy;
    }
    XFillPolygon(dsp, win, agc, points, npoints, Convex, CoordModeOrigin);
  }
  if (PWDAPPOXEQ(lisx, lisy, lisz, svi, liisx, liisy, liisz, svii, eps)) {
    return 1;
  }
  val           = val + delta;
  range         = ColorFringeValueToRange(ft, val);
  color         = ColorFringeRangeToColor(range);
  lower         = lower + delta;
  goto continue_1;
}



static int 
compute_poly_bbox(FPNum *x, FPNum *y, int npoints, EBoxVC2DRec *box)
{
  int i;
  
  box->left   = x[0];
  box->right  = x[0];
  box->bottom = y[0];
  box->top    = y[0];
  for (i = 1; i < npoints; i++) {
    box->left   = min(box->left, x[i]);
    box->right  = max(box->right, x[i]);
    box->bottom = min(box->bottom, y[i]);
    box->top    = max(box->top, y[i]);
  }
  return 1;
}



static int 
next_point_wd(FPNum sx, FPNum sy, FPNum sz,
              FPNum ex, FPNum ey, FPNum ez,
              FPNum sv, FPNum ev, FPNum lower, FPNum delta,
              FPNum *nx, FPNum *ny, FPNum *nz, FPNum *ns, FPNum *ne)
{
  double r;
#define EPS 1.0e-9

  if (fabs(ev - sv) < EPS ) {
    *nx = ex; *ny = ey; *nz = ez; *ns = ev; *ne = ev;
    return 1;
  }
  if (lower + delta <= ev) 
    *ns = lower + delta;
  else
    *ns = ev;

  r  = 1. / (ev - sv);
  r  *=  (*ns - sv);
  *nx = sx + r * (ex - sx);
  *ny = sy + r * (ey - sy);
  *nz = sz + r * (ez - sz);
  *ne = ev;
  return 1;
}



static void 
indices_seq(FPNum *values, int *indx1, int *indx2, int *indx3)
{
  if    (values[0] < values[1]) {
    if     (values[1] < values[2]) {
      *indx1 = 0; *indx2 = 1; *indx3 = 2;
    } else if (values[2] < values[0]) {
      *indx1 = 2; *indx2 = 0; *indx3 = 1;
    } else {
      *indx1 = 0; *indx2 = 2; *indx3 = 1;
    }
  }else {
    if     (values[0] < values[2]) {
      *indx1 = 1; *indx2 = 0; *indx3 = 2;
    } else if (values[2] < values[1]) {
      *indx1 = 2; *indx2 = 1; *indx3 = 0;
    } else {
      *indx1 = 1; *indx2 = 2; *indx3 = 0;
    }
  }
}

