
#include "Esimple.h" /* ELIXIR include  */
#include "ioP.h" /* Ckit include  */
#include "tokensP.h" /* Ckit include  */


#ifndef min
#    define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#    define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

void
make_from_dxf (char *geom_file_name);


static EPixel pix = 0;
static BOOLEAN status;
static EModel *elixir_model;
static FILE *handle = NULL;
static ckit_iobuf_t IOB = NULL;

static void
make_polygon (int nvert, WCRec poly[4])
{
  WCRec points[4];
  int i;
  GraphicObj *g_p;
#define ASSIGN_POINT_TO_POINT(DEST, SRC)                         \
  {                                                              \
     (DEST).x = (SRC).x; (DEST).y = (SRC).y; (DEST).z = (SRC).z; \
   }

  if        (nvert == 3) {
    for (i = 0; i < nvert; i++) {
      points[i].x = poly[i].x;
      points[i].y = poly[i].y;
      points[i].z = poly[i].z;
    }
    EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateTriangle3D (points));
    EGWithMaskChangeAttributes (COLOR_MASK | SHRINK_MASK | FILL_MASK, g_p);
  } else if (nvert == 4) {
    for (i = 0; i < nvert; i++) {
      points[i].x = poly[i].x;
      points[i].y = poly[i].y;
      points[i].z = poly[i].z;
    }
    EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateQuad3D(points));
    EGWithMaskChangeAttributes (COLOR_MASK | SHRINK_MASK | FILL_MASK, g_p);
  }  else {
    fprintf(stderr, "Cannot handle poly for nvert = %d\n", nvert);
  }
}


typedef struct box_t {
  WCRec         lower, upper;
}              box_t;
#define INIT_BOX(BOX, PT)			\
  {						\
    (BOX).lower.x = (PT).x;			\
    (BOX).lower.y = (PT).y;			\
    (BOX).lower.z = (PT).z;			\
    (BOX).upper.x = (PT).x;			\
    (BOX).upper.y = (PT).y;			\
    (BOX).upper.z = (PT).z;			\
  }


#define UPDT_BOX(BOX, PT)			\
  {						\
    (BOX).lower.x = min((BOX).lower.x, (PT).x);	\
    (BOX).lower.y = min((BOX).lower.y, (PT).y);	\
    (BOX).lower.z = min((BOX).lower.z, (PT).z);	\
    (BOX).upper.x = max((BOX).upper.x, (PT).x);	\
    (BOX).upper.y = max((BOX).upper.y, (PT).y);	\
    (BOX).upper.z = max((BOX).upper.z, (PT).z);	\
  }
#define BOX_RANGE(BOX, DIR) ((BOX).upper.DIR - (BOX).lower.DIR)

static int
how_many_verts (WCRec poly[4])
{
  box_t box;
  double tol;
  
  INIT_BOX (box, poly[0]);
  UPDT_BOX (box, poly[1]);
  UPDT_BOX (box, poly[2]);
  UPDT_BOX (box, poly[3]);

  tol = BOX_RANGE (box, x) + BOX_RANGE (box, y) + BOX_RANGE (box, z);
  tol /= 1000;

  if (   fabs (poly[2].x - poly[3].x) < tol
      && fabs (poly[2].y - poly[3].y) < tol
      && fabs (poly[2].z - poly[3].z) < tol)
    return 3;
  else
    return 4;
}


void
make_from_dxf (char *dxf_file_name)
{
  const char *buf, *tok;
  int readon = TRUE;
  static tokens_parser parser = NULL;
  WCRec poly[4];

  elixir_model = ESIModel();
  
  pix = EASValsGetColor();
  if (pix == (EPixel)0) {
    pix = ColorGetPixelFromString ("cyan", &status);
    if (!status) pix = 0;
  }
  EASValsSetColor (pix); 
  EASValsSetShrink (1.0);
  EASValsSetFillStyle (FILL_SOLID);
  
  handle = ckit_io_open_stream (dxf_file_name, "r");
  IOB    = ckit_io_new_buffer (handle);

  if (parser == NULL) {
    parser = tokens_new_parser ();
  }
  
  while (readon) {
    buf = ckit_io_scan_line (IOB, FALSE);
    if (buf == NULL) {
      goto exit;
    }
    tokens_parse_line (parser, (char *)buf);
    if (tokens_type_of_token (parser, 1) == TOKENS_STRING) {
      tok = tokens_token_as_string (parser, 1);
      if        (tokens_token_and_keyword_equiv (tok, "3DFACE")) {
#       define SKIP_LINE() ckit_io_scan_line (IOB, FALSE)
#       define CHKBUF                                   \
        {                                               \
          if (buf == NULL) {                            \
            fprintf (stderr,                            \
                     "Expected data, l.%d\n",           \
                     ckit_io_line_number (IOB));        \
            goto exit;                                  \
          }                                             \
        }
#       define GETBUF                                   \
        {                                               \
          buf = ckit_io_scan_line (IOB, FALSE);         \
          CHKBUF;                                       \
          tokens_parse_line (parser, (char *)buf);      \
        }
#       define GETCODE(gc)                                              \
        {                                                               \
          gc = 999;                                                     \
          GETBUF;                                                       \
          if (tokens_type_of_token (parser, 1) != TOKENS_NUMBER) {      \
            fprintf (stderr,                                            \
                     "Expected group code, l.%d, got \"%s\"\n",         \
                     ckit_io_line_number (IOB),                         \
                     ckit_io_current_line (IOB, FALSE));                \
          }                                                             \
          gc = tokens_token_as_int (parser, 1);                         \
        }
        int nvert;
        int gc;
        int got_nvals;

        got_nvals = 0;
        while (got_nvals != 12) {
          GETCODE (gc);
          switch (gc) {
          case 10: case 11: case 12: case 13:
            GETBUF;
            poly[gc - 10].x = tokens_token_as_double (parser, 1); got_nvals++;
            break;
          case 20: case 21: case 22: case 23:
            GETBUF;
            poly[gc - 20].y = tokens_token_as_double (parser, 1); got_nvals++;
            break;
          case 30: case 31: case 32: case 33:
            GETBUF;
            poly[gc - 30].z = tokens_token_as_double (parser, 1); got_nvals++;
            break;
          default:
            SKIP_LINE ();
            break;
          }
        }
        nvert = how_many_verts (poly);
        make_polygon (nvert, poly);
      } else if (tokens_token_and_keyword_equiv (tok, "EOF")) {
        readon = FALSE;
      } else {
        /* ignore */
      }
    }
  skip:;
  } /* while (readon) */
  
 exit:
  ckit_io_delete_buffer (IOB);
  fclose(handle);
}





