
#include "Esimple.h" /* ELIXIR include  */
#include "ioP.h" /* Ckit include  */
#include "tokensP.h" /* Ckit include  */


#ifndef min
#    define min(a,b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef max
#    define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

void
make_from_fct (char *geom_file_name);


static EPixel pix = 0;
static BOOLEAN status;
static EModel *elixir_model;
static FILE *handle = NULL;
static ckit_iobuf_t IOB = NULL;

static void
make_polygon (int nvert, WCRec *poly)
{
  WCRec points[4];
  int i;
  GraphicObj *g_p;
#define ASSIGN_POINT_TO_POINT(DEST, SRC)                         \
  {                                                              \
     (DEST).x = (SRC).x; (DEST).y = (SRC).y; (DEST).z = (SRC).z; \
   }

  if        (nvert == 3) {
    for (i = 0; i < nvert; i++) {
      points[i].x = poly[i].x;
      points[i].y = poly[i].y;
      points[i].z = poly[i].z;
    }
    EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateTriangle3D (points));
    EGWithMaskChangeAttributes (COLOR_MASK | SHRINK_MASK | FILL_MASK, g_p);
  } else if (nvert == 4) {
    for (i = 0; i < nvert; i++) {
      points[i].x = poly[i].x;
      points[i].y = poly[i].y;
      points[i].z = poly[i].z;
    }
    EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateQuad3D(points));
    EGWithMaskChangeAttributes (COLOR_MASK | SHRINK_MASK | FILL_MASK, g_p);
  }  else {
    fprintf(stderr, "Cannot handle poly for nvert = %d\n", nvert);
  }
}


typedef struct box_t {
  WCRec         lower, upper;
}              box_t;
#define INIT_BOX(BOX, PT)			\
  {						\
    (BOX).lower.x = (PT).x;			\
    (BOX).lower.y = (PT).y;			\
    (BOX).lower.z = (PT).z;			\
    (BOX).upper.x = (PT).x;			\
    (BOX).upper.y = (PT).y;			\
    (BOX).upper.z = (PT).z;			\
  }


#define UPDT_BOX(BOX, PT)			\
  {						\
    (BOX).lower.x = min((BOX).lower.x, (PT).x);	\
    (BOX).lower.y = min((BOX).lower.y, (PT).y);	\
    (BOX).lower.z = min((BOX).lower.z, (PT).z);	\
    (BOX).upper.x = max((BOX).upper.x, (PT).x);	\
    (BOX).upper.y = max((BOX).upper.y, (PT).y);	\
    (BOX).upper.z = max((BOX).upper.z, (PT).z);	\
  }
#define BOX_RANGE(BOX, DIR) ((BOX).upper.DIR - (BOX).lower.DIR)

static int
how_many_verts (WCRec poly[4])
{
  box_t box;
  double tol;
  
  INIT_BOX (box, poly[0]);
  UPDT_BOX (box, poly[1]);
  UPDT_BOX (box, poly[2]);
  UPDT_BOX (box, poly[3]);

  tol = BOX_RANGE (box, x) + BOX_RANGE (box, y) + BOX_RANGE (box, z);
  tol /= 1000;

  if (   fabs (poly[2].x - poly[3].x) < tol
      && fabs (poly[2].y - poly[3].y) < tol
      && fabs (poly[2].z - poly[3].z) < tol)
    return 3;
  else
    return 4;
}


void
make_from_fct (char *dxf_file_name)
{
  const char *buf, *tok;
  int readon = TRUE;
  static tokens_parser parser = NULL;
  int npt, tpt;
  WCRec *pt;
#define CHKBUF                                          \
        {                                               \
          if (buf == NULL) {                            \
            fprintf (stderr,                            \
                     "Expected data, l.%d\n",           \
                     ckit_io_line_number (IOB));        \
            goto exit;                                  \
          }                                             \
        }
#define GETBUF                                          \
        {                                               \
          buf = ckit_io_scan_line (IOB, FALSE);         \
          CHKBUF;                                       \
          tokens_parse_line (parser, (char *)buf);      \
        }
#define REALLOC_PT_BUF(npt)                             \
  {                                                     \
    if (!pt || npt+1 > tpt) {                           \
      tpt = 2*npt + 1;                                  \
      pt = NEWP_REALLOC_OBJ_ARRAY (pt, tpt, WCRec);     \
      if (!pt) {                                        \
        fprintf (stderr, "Not enough memory\n");        \
        goto exit;                                      \
      }                                                 \
    }                                                   \
  }
  
  elixir_model = ESIModel();
  
  pix = EASValsGetColor();
  if (pix == (EPixel)0) {
    pix = ColorGetPixelFromString ("cyan", &status);
    if (!status) pix = 0;
  }
  EASValsSetColor (pix); 
  EASValsSetShrink (1.0);
  EASValsSetFillStyle (FILL_SOLID);
  
  handle = ckit_io_open_stream (dxf_file_name, "r");
  IOB    = ckit_io_new_buffer (handle);

  if (parser == NULL) {
    parser = tokens_new_parser ();
  }

  GETBUF;
  if (tokens_type_of_token (parser, 1) == TOKENS_STRING) {
    tok = tokens_token_as_string (parser, 1);
    if        (!tokens_token_and_keyword_equiv (tok, "PPM_START")) {
      fprintf (stderr, "Expected \"PPM_START\"");
      goto exit;
    }
  } else {
    fprintf (stderr, "Expected \"PPM_START\", l.%d, got \"%s\"\n",  
             ckit_io_line_number (IOB),                         
             ckit_io_current_line (IOB, FALSE));
  }

  tpt = 0; pt  = NULL;

  while (readon) {
    GETBUF;
    if (tokens_type_of_token (parser, 1) == TOKENS_STRING) {
      tok = tokens_token_as_string (parser, 1);
      if        (tokens_token_and_keyword_equiv (tok, "POLY_START")) {
        npt = 0;
        while (1) {
          GETBUF;
          if (tokens_type_of_token (parser, 1) == TOKENS_STRING) {
            tok = tokens_token_as_string (parser, 1);
            if        (tokens_token_and_keyword_equiv (tok, "POSITION:")) {
              REALLOC_PT_BUF (npt);
              if (  (tokens_total_of_tokens (parser) != 4)
                  || tokens_type_of_token (parser, 2) != TOKENS_NUMBER
                  || tokens_type_of_token (parser, 3) != TOKENS_NUMBER
                  || tokens_type_of_token (parser, 4) != TOKENS_NUMBER
                  ) {
                fprintf (stderr,
                         "Expected three numbers, l.%d, got \"%s\"\n",  
                         ckit_io_line_number (IOB),                         
                         ckit_io_current_line (IOB, FALSE));
                goto exit;
              }
              pt[npt].x = tokens_token_as_double (parser, 2);
              pt[npt].y = tokens_token_as_double (parser, 3);
              pt[npt].z = tokens_token_as_double (parser, 4);
              npt++;
            } else if (tokens_token_and_keyword_equiv (tok, "POLY_END")) {
              make_polygon (npt, pt);
              break;
            } else {
              fprintf (stderr, "Expected \"POLY_END\" or \"POSITION:\","
                       "l.%d, got \"%s\"\n",  
                       ckit_io_line_number (IOB),                         
                       ckit_io_current_line (IOB, FALSE));
            }
          } else {
            fprintf (stderr, "Expected \"POLY_END\" or \"POSITION:\","
                     "l.%d, got \"%s\"\n",  
                     ckit_io_line_number (IOB),                         
                     ckit_io_current_line (IOB, FALSE));
            goto exit;
          }
        } /* while */
      } else if (tokens_token_and_keyword_equiv (tok, "PPM_END")) {
        readon = FALSE;
      } else {
        fprintf (stderr, "Unknown input, l.%d, got \"%s\"\n",  
                     ckit_io_line_number (IOB),                         
                     ckit_io_current_line (IOB, FALSE));
      }
    }
  skip:;
  } /* while (readon) */
  
 exit:
  ckit_io_delete_buffer (IOB);
  fclose(handle);
}





