
#include "Esimple.h" /* ELIXIR include  */
#include "ioP.h" /* Ckit include  */
#include "tokensP.h" /* Ckit include  */

#include <math.h>
#include <stdio.h>

static void
make_vertex(int i, WCRec *p);
static void
make_polygon(int vert, WCRec poly[]);
void
make_from_geom(char *geom_file_name);


static EPixel pix = 0;
static BOOLEAN status;
static EModel *elixir_model;
static FILE *handle = NULL;
static ckit_iobuf_t IOB = NULL;

static int                             
tokens_get_token (char *line,    /* input line */
                  int *char_p,   /* index the current token; */
                  /* must be set *char_p=0 before the first call */
                  char token[]  /* token as string */
                  )
{
  int	tok_pos;

#define SPACE	 ' '
#define TAB	 '\t'
#define DOT      '.'
#define PLUS     '+'
#define MINUS    '-'
#define MINUS    '-'
#define EOS      '\0'
#define UNDERBAR '_'
#define NEWLINE  '\n'
#define DQUOTE   '"'

  
  token[0] = EOS;
  tok_pos = 0;

  if (line[*char_p] == EOS || line[*char_p] == NEWLINE)
    return (0);
  /*  Either remove runs of whitespace  */
  /*  or return the stuff  */
  while (line[*char_p] == SPACE || line[*char_p] == TAB) {
    (*char_p)++;
    if (line[*char_p] == EOS || line[*char_p] == NEWLINE)
      return (0);
  }
  
  /*  Treat quoted strings here  */
  if (line[*char_p] == DQUOTE) {
    (*char_p)++;
    while (line[*char_p] != EOS && line[*char_p] != NEWLINE) {
      token[tok_pos] = line[(*char_p)++];
      /*  Check for doubled up quotes  */
      if (token[tok_pos] != DQUOTE) {
        tok_pos++;
      } else {
        if (line[*char_p] == DQUOTE) {
          tok_pos++;
          (*char_p)++;
        } else {
          token[tok_pos] = EOS;
          return (1);
        }
      }
    }
    token[tok_pos] = EOS;
    return (1);
  }

  /*  Treat any other input here  */
  if (!isspace (line[*char_p])) {
    while (line[*char_p] != EOS && line[*char_p] != NEWLINE) {
      token[tok_pos] = line[*char_p];
      if (isspace (line[*char_p])) {
        token[tok_pos] = EOS;
        return (1);
      }
      tok_pos++;
      (*char_p)++;
    }
    token[tok_pos] = EOS;
    return (1);
  }

  return (0);
}


void
make_from_geom(char *geom_file_name)
{
  WCRec poly[100];
  WCRec *pts;
  int num_vec,num_poly,i,j,vert,value;
  const char *buf;
  char base[5];
  int base_index;

  elixir_model = ESIModel();
  
  pix = EASValsGetColor();
  if (pix == (EPixel)0) {
    pix = ColorGetPixelFromString("cyan", &status);
    if (!status) pix = 0;
  }
  EASValsSetColor(pix);
  EASValsSetShrink(1.0);
  EASValsSetFillStyle(FILL_SOLID);
  
  handle = ckit_io_open_stream (geom_file_name, "r");
  IOB    = ckit_io_new_buffer (handle);

 new_data:
  buf = ckit_io_scan_line (IOB, 0);
  if (buf == NULL) goto exit;
  base_index = 1;
  if (sscanf(buf, "%d %d %s", &num_vec, &num_poly, base) == 3) {
    if (strcmp(base, "zero") == 0) base_index = 0;
    else                           base_index = 1;
  } else if (sscanf(buf, "%d %d", &num_vec, &num_poly) != 2)
    goto exit;

#if 0
  printf("Geom data from %s: %d vertices, %d polygons\n", geom_file_name,
         num_vec, num_poly);
#endif

  pts = NEWP_OBJ_ARRAY (num_vec, WCRec);
  if (!pts) {
    fprintf(stderr, "Not enough memory to load data\n");
    goto exit;
  }
  
  for(i=0;i<num_vec;i++) {
    float x, y, z;
    buf = ckit_io_scan_line (IOB, 0);
    if (buf == NULL) goto exit;
    if (sscanf(buf, "%f %f %f", &x, &y, &z) != 3)
      fprintf(stderr, "Bad vertex data in line %d\n",
              ckit_io_line_number (IOB));
    pts[i].x = x;
    pts[i].y = z; /* NOTE that y and z are swapped here */
    pts[i].z = y;
    make_vertex(i, &pts[i]);
  }
  for(i=0;i<num_poly;i++) {
    int tok_type;
    int ptr = 0;
    char tok[256];
    
    buf = ckit_io_scan_line (IOB, 0);
    if (buf == NULL) goto exit;
    if (sscanf(buf,"%d", &vert) != 1)
      fprintf(stderr, "Bad poly data in line %d\n", ckit_io_line_number (IOB));
    j = -1;
    while ((tok_type = tokens_get_token((char *) buf, &ptr, tok))
           && j < vert) {
      if (sscanf(tok,"%d", &value) != 1)
        fprintf(stderr, "Bad poly data in line %d\n",
                ckit_io_line_number (IOB));
      if (j >= 0) {/* skip j=-1 -- that is the /vert/ data */
        if (base_index == 0) value++;
        if (base_index == 1 && value == 0) {
          fprintf(stderr, "Zero index in line %d while base_index == 1;\n   "
                  "set zero base index\n", ckit_io_line_number (IOB));
          exit(1);
        }
        poly[j].x = pts[value-1].x;   
        poly[j].y = pts[value-1].y;   
        poly[j].z = pts[value-1].z;
      } 
      j++;
    }
    make_polygon(vert, poly);
  }
  
#if 0
  printf("Done.\n\r");
#endif
  NEWP_FREE_OBJ (pts);
  
  goto new_data;

  exit:
  ckit_io_delete_buffer (IOB);
  fclose(handle);
}



#define ASSIGN_POINT_TO_POINT(DEST, SRC)                         \
  {                                                              \
     (DEST).x = (SRC).x; (DEST).y = (SRC).y; (DEST).z = (SRC).z; \
   }

static void
make_polygon(int vert, WCRec poly[6])
{
  WCRec points[6];
  int i;
  GraphicObj *g_p;
  
  if        (vert == 3) {
    for (i = 0; i < vert; i++) {
      points[i].x = poly[i].x;
      points[i].y = poly[i].y;
      points[i].z = poly[i].z;
    }
    EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateTriangle3D(points));
    EGWithMaskChangeAttributes(COLOR_MASK
                               | SHRINK_MASK
                               | FILL_MASK, g_p);
  } else if (vert == 4) {
    for (i = 0; i < vert; i++) {
      points[i].x = poly[i].x;
      points[i].y = poly[i].y;
      points[i].z = poly[i].z;
    }
    EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateQuad3D(points));
    EGWithMaskChangeAttributes(COLOR_MASK
                               | SHRINK_MASK
                               | FILL_MASK, g_p);
  }  else if (vert == 5) {
    ASSIGN_POINT_TO_POINT(points[0], poly[0]);
    ASSIGN_POINT_TO_POINT(points[1], poly[1]);
    ASSIGN_POINT_TO_POINT(points[2], poly[2]);
    EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateTriangle3D(points));
    EGWithMaskChangeAttributes(COLOR_MASK
                               | SHRINK_MASK
                               | FILL_MASK, g_p);
    ASSIGN_POINT_TO_POINT(points[0], poly[0]);
    ASSIGN_POINT_TO_POINT(points[1], poly[2]);
    ASSIGN_POINT_TO_POINT(points[2], poly[3]);
    EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateTriangle3D(points));
    EGWithMaskChangeAttributes(COLOR_MASK
                               | SHRINK_MASK
                               | FILL_MASK, g_p);
    ASSIGN_POINT_TO_POINT(points[0], poly[0]);
    ASSIGN_POINT_TO_POINT(points[1], poly[3]);
    ASSIGN_POINT_TO_POINT(points[2], poly[4]);
    EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateTriangle3D(points));
    EGWithMaskChangeAttributes(COLOR_MASK
                               | SHRINK_MASK
                               | FILL_MASK, g_p);
    }  else if (vert == 6) {
    ASSIGN_POINT_TO_POINT(points[0], poly[0]);
    ASSIGN_POINT_TO_POINT(points[1], poly[1]);
    ASSIGN_POINT_TO_POINT(points[2], poly[5]);
    EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateTriangle3D(points));
    EGWithMaskChangeAttributes(COLOR_MASK
                               | SHRINK_MASK
                               | FILL_MASK, g_p);
    ASSIGN_POINT_TO_POINT(points[0], poly[4]);
    ASSIGN_POINT_TO_POINT(points[1], poly[2]);
    ASSIGN_POINT_TO_POINT(points[2], poly[3]);
    EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateTriangle3D(points));
    EGWithMaskChangeAttributes(COLOR_MASK
                               | SHRINK_MASK
                               | FILL_MASK, g_p);
    ASSIGN_POINT_TO_POINT(points[0], poly[1]);
    ASSIGN_POINT_TO_POINT(points[1], poly[2]);
    ASSIGN_POINT_TO_POINT(points[2], poly[5]);
    EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateTriangle3D(points));
    EGWithMaskChangeAttributes(COLOR_MASK
                               | SHRINK_MASK
                               | FILL_MASK, g_p);
    ASSIGN_POINT_TO_POINT(points[0], poly[4]);
    ASSIGN_POINT_TO_POINT(points[1], poly[2]);
    ASSIGN_POINT_TO_POINT(points[2], poly[5]);
    EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateTriangle3D(points));
    EGWithMaskChangeAttributes(COLOR_MASK
                               | SHRINK_MASK
                               | FILL_MASK, g_p);
  } else {
    fprintf(stderr, "Cannot handle poly for vert = %d\n", vert);
  }
}


static void
make_vertex(int i, WCRec  *p)
{
#if 0
  WCRec pnt;
  char buf[24];
  GraphicObj *g_p;

  pnt.x = p->x;
  pnt.y = p->y;
  pnt.z = p->z;
  sprintf(buf, "%-d", i+1);
  EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateAnnText3D(&pnt, buf));
  EGWithMaskChangeAttributes(COLOR_MASK, g_p);
#endif
}


