
#include "Esimple.h" /* ELIXIR include  */
#include "ioP.h" /* Ckit include  */
#include "tokensP.h" /* Ckit include  */

#include <math.h>
#include <stdio.h>


static EPixel fpix = 0;
static EPixel epix = 0;
static BOOLEAN status;
static EModel *elixir_model;

  typedef struct e_t {
    int n[2];
  } e_t;

  typedef struct v_t {
    double x, y, z;
  } v_t;

  typedef struct t_t {
    int e[3];
  } t_t;

static int
common_v (e_t *e1, e_t *e2)
{
  if      (e1->n[0] == e2->n[0]) return e1->n[0];
  else if (e1->n[1] == e2->n[0]) return e1->n[1];
  else if (e1->n[0] == e2->n[1]) return e1->n[0];
  else if (e1->n[1] == e2->n[1]) return e1->n[1];
  else                           return -1;
}
  
static void
vns (t_t *t, e_t *es, int vn[3])
{
  vn[0] = common_v (&es[t->e[0]-1], &es[t->e[1]-1]);
  vn[1] = common_v (&es[t->e[1]-1], &es[t->e[2]-1]);
  vn[2] = common_v (&es[t->e[2]-1], &es[t->e[0]-1]);
}

void
make_from_gts (char *file_name)
{
  e_t *es = 0;
  v_t *vs = 0;
  t_t *ts = 0;
  FILE *fp = 0;
  int i, nv = 0, ne = 0, nt = 0;
  int line = 1;

  fp = fopen(file_name, "r");

  {
    const char *buf;
    ckit_iobuf_t iob = ckit_io_new_buffer (fp);
    buf = ckit_io_scan_line (iob, 0);
    {
      tokens_parser p = tokens_new_parser();
      if (!tokens_parse_line (p, buf)) {
        fprintf(stderr, "Bad totals in line 1\n");
        goto exit;
      }
      if (tokens_total_of_tokens (p) < 3) {
        fprintf(stderr, "Bad totals in line 1\n");
        goto exit;
      }
      nv = tokens_token_as_int (p, 1);
      ne = tokens_token_as_int (p, 2);
      nt = tokens_token_as_int (p, 3);
      tokens_delete_parser(p);
    }
    ckit_io_delete_buffer (iob);
  }
  line++;
  
  vs =  NEWP_OBJ_ARRAY (nv, v_t);
  if (!vs) {
    fprintf(stderr, "Not enough memory to load data\n");
    goto exit;
  }
  es =  NEWP_OBJ_ARRAY (ne, e_t);
  if (!es) {
    fprintf(stderr, "Not enough memory to load data\n");
    goto exit;
  }
  ts =  NEWP_OBJ_ARRAY (nt, t_t);
  if (!ts) {
    fprintf(stderr, "Not enough memory to load data\n");
    goto exit;
  }
 
  for(i=0;i<nv;i++) {
    if (fscanf(fp, "%lf %lf %lf", &vs[i].x, &vs[i].y, &vs[i].z) != 3) {
      fprintf(stderr, "Bad vertex data in line %d\n", line);
      goto exit;
    }
    line++;
  }
  
  for(i=0;i<ne;i++) {
    if (fscanf(fp, "%d %d", &es[i].n[0], &es[i].n[1]) != 2) {
      fprintf(stderr, "Bad edge data in line %d\n", line);
      goto exit;
    }
    line++;
  }
  
  for(i=0;i<nt;i++) {
    if (fscanf(fp, "%d %d %d", &ts[i].e[0], &ts[i].e[1], &ts[i].e[2]) != 3) {
      fprintf(stderr, "Bad triangle data in line %d\n", line);
      goto exit;
    }
    line++;
  }
  
  elixir_model = ESIModel();
  
  fpix = EASValsGetColor();
  if (fpix == (EPixel)0) {
    fpix = ColorGetPixelFromString("cyan", &status);
    if (!status) fpix = 0;
    epix = ColorGetPixelFromString("black", &status);
  }
  EASValsSetColor(fpix);
  EASValsSetShrink(1.0);
  EASValsSetEdgeFlag(1);
  EASValsSetEdgeColor(epix);
  EASValsSetFillStyle(FILL_SOLID);

  for (i = 0; i < nt; i++) {
    {
      WCRec points[3];
      int m;
      GraphicObj *g_p;
      int vn[3];
      vns(&ts[i], es, vn);
      for (m = 0; m < 3; m++) {
        int nid = vn[m]-1;
        if (nid < 0) {
          fprintf(stderr, "Bad connectivity data: %d\n", nid);
          goto exit;
        }
        points[m].x = vs[nid].x;
        points[m].y = vs[nid].y;
        points[m].z = vs[nid].z;
      }
      EMNoDrawAddGraphicsToModel (elixir_model, g_p = CreateTriangle3D(points));
      EGWithMaskChangeAttributes(COLOR_MASK
                                 | EDGE_COLOR_MASK
                                 | EDGE_FLAG_MASK
                                 | SHRINK_MASK
                                 | FILL_MASK, g_p);
    }
  }
  
  exit:
  if (vs) NEWP_FREE_OBJ (vs);
  if (es) NEWP_FREE_OBJ (es);
  if (ts) NEWP_FREE_OBJ (ts);
  fclose(fp);
}



