/*
 * Copyright 1989 Software Research Associates, Inc., Tokyo, Japan
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be used
 * in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  Software Research Associates
 * makes no representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 *
 * SOFTWARE RESEARCH ASSOCIATES DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL SOFTWARE RESEARCH ASSOCIATES BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik M. van der Poel
 *         Software Research Associates, Inc., Tokyo, Japan
 *         erik@sra.co.jp
 */

/*
 * Author's address:
 *
 *     erik@sra.co.jp
 *                                            OR
 *     erik%sra.co.jp@uunet.uu.net
 *                                            OR
 *     erik%sra.co.jp@mcvax.uucp
 *                                            OR
 *     try junet instead of co.jp
 *                                            OR
 *     Erik M. van der Poel
 *     Software Research Associates, Inc.
 *     1-1-1 Hirakawa-cho, Chiyoda-ku
 *     Tokyo 102 Japan. TEL +81-3-234-2692
 */

#include "Econfig.h"
#include "Ecfg.h"
#include <stdio.h>
#include <errno.h>

#include "Eselfile.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif /* ndef MAXPATHLEN */

#ifdef HAVE_GETWD
extern char *getwd();
#endif 

Widget
ESITopLevelWidget(void);

static int SFstatus = SEL_FILE_NULL;

static char
	SFstartDir[MAXPATHLEN],
	SFcurrentPath[MAXPATHLEN],
	SFcurrentDir[MAXPATHLEN];

static Widget
	selFile,
	selFileCancel,
	selFileField,
	selFileForm,
	selFileHScroll,
	selFileHScrolls[3],
	selFileLists[3],
	selFileOK,
	selFilePrompt,
	selFileVScrolls[3];

static Display *SFdisplay;

static Pixel SFfore, SFback;

static Atom	SFwmDeleteWindow;

static XSegment SFsegs[2], SFcompletionSegs[2];

static XawTextPosition SFtextPos;

static int SFupperX, SFlowerY, SFupperY;

static int SFtextX, SFtextYoffset;

static int SFentryWidth, SFentryHeight;

static int SFlineToTextH = 3;

static int SFlineToTextV = 3;

static int SFbesideText = 3;

static int SFaboveAndBelowText = 2;

static int SFcharsPerEntry = 15;

static int SFlistSize = 10;

static int SFworkProcAdded = 0;

static XtAppContext SFapp;

static int SFpathScrollWidth, SFvScrollHeight, SFhScrollWidth;

static char SFtextBuffer[MAXPATHLEN];

static XtIntervalId SFdirModTimerId;

int (*SFfunc)();

static int SFcharWidth, SFcharAscent, SFcharHeight;

static int SFcurrentInvert[3] = { -1, -1, -1 };
 
static GC SFlineGC, SFscrollGC, SFinvertGC, SFtextGC;

typedef struct {
	char	*name;
	char	*dir;
} SFLogin;

static int
SFfree(int i);
static int
SFstrdup(char	**s1, char * s2);
static int
SFunreadableDir(SFDir	*dir);
#ifdef SEL_FILE_IGNORE_CASE
static int
SFstrncmp(register char	*p, register char *q, register int n);
#endif
static int
SFreplaceText(SFDir	*dir, char	*str);
static void
SFexpand(char	*str);
static int
SFfindFile(SFDir		*dir, register char	*str);
static int
SFunselect(void);
static int
SFcompareLogins(SFLogin	*p, SFLogin	*q);
static int
SFgetHomeDirs(void);
static int
SFfindHomeDir(char	*begin, char	*end);
static int
SFcheckDir(int n, SFDir		*dir);
static int
SFcheckFiles(SFDir	*dir);
static int
SFtextChanged(void);

static SFDir *SFdirss = NULL;

static int SFdirEnd;

static int SFdirPtr;

static int SFbuttonPressed = 0;

static int SFdoNotTouchDirPtr = 0;

static int SFdoNotTouchVorigin = 0;

static SFDir SFrootDir, SFhomeDir;

static SFLogin *SFlogins;

static int SFtwiddle = 0;

static char *oneLineTextEditTranslations = "\
	<Key>Return:	redraw-display()\n\
	Ctrl<Key>M:	redraw-display()\n\
";

#define SF_DEFAULT_FONT "9x15"

#ifdef ABS
#undef ABS
#endif
#define ABS(x) (((x) < 0) ? (-(x)) : (x))

typedef struct {
	char *fontname;
} TextData, *textPtr;


static XtResource textResources[] = {
	{XtNfont, XtCFont, XtRString, sizeof (char *),
		XtOffset(textPtr, fontname), XtRString, SF_DEFAULT_FONT},
};

static XFontStruct *SFfont;

static int SFcurrentListY;

static XtIntervalId SFscrollTimerId;

static int
SFdrawStrings(register Window w, register SFDir *dir,
              register int intfrom, register int intto);
static int
SFwriteStatChar(char *name, int last, struct stat *statBuf);
static int
SFdrawStrings(register Window w, register SFDir *dir,
              register int intfrom, register int intto);
static int
SFinvertEntry(int n);
static int
SFdrawList(int n, int doScroll);
static int
SFdrawLists(int doScroll);

static unsigned long
SFscrollTimerInterval(void);
static void
SFscrollTimer(XtPointer p, XtIntervalId *id);
static int
SFnewInvertEntry(register int n, register XButtonReleasedEvent *event);

static void
SFhSliderMovedCallback(Widget w, int n, float *new);

/*
 * Copyright 1989 Software Research Associates, Inc., Tokyo, Japan
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be used
 * in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  Software Research Associates
 * makes no representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 *
 * SOFTWARE RESEARCH ASSOCIATES DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL SOFTWARE RESEARCH ASSOCIATES BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik M. van der Poel
 *         Software Research Associates, Inc., Tokyo, Japan
 *         erik@sra.co.jp
 */

#include <stdio.h>

#ifdef SEL_FILE_IGNORE_CASE
#include <ctype.h>
#endif /* def SEL_FILE_IGNORE_CASE */

#include "Eselfile.h"
#include "Ecfg.h"

#if HAVE_DIRENT_H
#  include <dirent.h>
#  define NAMLEN(dirent) strlen((dirent)->d_name)
#  define NAME(dirent)   (dirent)->d_name
#else 
#  define dirent direct
#  define NAMLEN(dirent) (dirent)->d_namlen
#  define NAME(dirent)   (dirent)->d_name
#  if HAVE_SYS_NDIR_H
#    include <sys/ndir.h>
#  endif
#  if HAVE_SYS_DIR_H
#    include <sys/dir.h>
#  endif
#  if HAVE_NDIR_H
#    include <ndir.h>
#  endif
#endif 

#include <sys/stat.h>

#if HAVE_QSORT
extern void qsort();
#endif 

#ifdef SEL_FILE_IGNORE_CASE
static int
SFcompareEntries(SFEntry *p, SFEntry *q)
{
  register char	*r, *s;
  register char	c1, c2;
  
  r = p->real;
  s = q->real;
  
  c1 = *r++;
  if (islower(c1)) {
    c1 = toupper(c1);
  }
  c2 = *s++;
  if (islower(c2)) {
    c2 = toupper(c2);
  }
  
  while (c1 == c2) {
    if (!c1) {
      return strcmp(p->real, q->real);
    }
    c1 = *r++;
    if (islower(c1)) {
      c1 = toupper(c1);
    }
    c2 = *s++;
    if (islower(c2)) {
      c2 = toupper(c2);
    }
  }
  
  return c1 - c2;
}
#else /* def SEL_FILE_IGNORE_CASE */
static int
SFcompareEntries(SFEntry *p, SFEntry *q)
{
  return strcmp(p->real, q->real);
}
#endif /* def SEL_FILE_IGNORE_CASE */

static int
SFgetDir(SFDir	*dir)
{
  SFEntry		*result = NULL;
  int		alloc = 0;
  int		i;
  DIR		*dirp;
  struct dirent	*dp;
  char		*str;
  int		len;
  int		maxChars;
  struct stat	statBuf;
  
  maxChars = strlen(dir->dir) - 1;
  
  dir->entries = NULL;
  dir->nEntries = 0;
  dir->nChars = 0;
  
  result = NULL;
  i = 0;
  
  dirp = opendir(".");
  if (!dirp) {
    return 1;
  }
  
  (void) stat(".", &statBuf);
  dir->mtime = statBuf.st_mtime;
  

  (void) readdir(dirp);	/* throw away "." */
  
#ifndef S_IFLNK
  (void) readdir(dirp);	/* throw away ".." */
#endif /* ndef S_IFLNK */
  
  while ((dp = readdir(dirp))) {
    if (i >= alloc) {
      alloc = 2 * (alloc + 1);
      result = (SFEntry *) XtRealloc((char *) result,
                                     (unsigned) (alloc * sizeof(SFEntry)));
    }
    result[i].statDone = 0;
    len = NAMLEN(dp);
    str = NAME(dp);
    result[i].real = XtMalloc((unsigned) (len + 2));
    (void) strcat(strcpy(result[i].real, str), " ");
    if (len > maxChars) {
      maxChars = len;
    }
    result[i].shown = result[i].real;
    i++;
  }
  
#if defined(SVR4) || defined(SYSV) || defined(USG)
  qsort((char *) result, (unsigned) i, sizeof(SFEntry), SFcompareEntries);
#else /* defined(SVR4) || defined(SYSV) || defined(USG) */
  qsort((char *) result, i, sizeof(SFEntry), SFcompareEntries);
#endif /* defined(SVR4) || defined(SYSV) || defined(USG) */
  
  dir->entries = result;
  dir->nEntries = i;
  dir->nChars = maxChars + 1;
  
  closedir(dirp);
  
  return 0;
}

static int
SFprepareToReturn(void);

static int
SFchdir(char	*path)
{
  int	result;
  
  result = 0;
  
  if (strcmp(path, SFcurrentDir)) {
    result = chdir(path);
    if (!result) {
      (void) strcpy(SFcurrentDir, path);
    }
  }
  
  return result;
}

static int
SFfree(int i)
{
  register SFDir	*dir;
  register int	j;
  
  dir = &(SFdirss[i]);
  
  for (j = dir->nEntries - 1; j >= 0; j--) {
    if (dir->entries[j].shown != dir->entries[j].real) {
      XtFree(dir->entries[j].shown);
    }
    XtFree(dir->entries[j].real);
  }
  
  XtFree((char *) dir->entries);
  
  XtFree(dir->dir);
  
  dir->dir = NULL;
  return 1;
}

static int
SFstrdup(char	**s1, char * s2)
{
  *s1 = strcpy(XtMalloc((unsigned) (strlen(s2) + 1)), s2);
  return 1;
}

static int
SFunreadableDir(SFDir	*dir)
{
  char	*cannotOpen = "<cannot open> ";
  
  dir->entries = (SFEntry *) XtMalloc(sizeof(SFEntry));
  dir->entries[0].statDone = 1;
  SFstrdup(&dir->entries[0].real, cannotOpen);
  dir->entries[0].shown = dir->entries[0].real;
  dir->nEntries = 1;
  dir->nChars = strlen(cannotOpen);
  return 1;
}

/* ARGSUSED */
static void
SFexposeList(Widget w, XtPointer n, XEvent *event, Boolean *cont)
{
  if ((event->type == NoExpose) || event->xexpose.count) {
    return;
  }
  
  SFdrawList((int)n, SF_DO_NOT_SCROLL);
}

/* ARGSUSED */
static void
SFmodVerifyCallback(Widget w, XtPointer client_data,
                    XEvent *event, Boolean *cont)
{
  char	buf[2];
  
  if (
      (XLookupString(&(event->xkey), buf, 2, NULL, NULL) == 1) &&
      ((*buf) == '\r')
      ) {
    SFstatus = SEL_FILE_OK;
  } else {
    SFstatus = SEL_FILE_TEXT;
  }
}

/* ARGSUSED */
static void
SFokCallback(Widget w, XtPointer cl, XtPointer cd)
{
  SFstatus = SEL_FILE_OK;
}

static XtCallbackRec SFokSelect[] = {
	{ SFokCallback, (XtPointer) NULL },
	{ NULL, (XtPointer) NULL },
};

/* ARGSUSED */
static void
SFcancelCallback(Widget w, XtPointer cl, XtPointer cd)
{
  SFstatus = SEL_FILE_CANCEL;
}

static XtCallbackRec SFcancelSelect[] = {
	{ SFcancelCallback, (XtPointer) NULL },
	{ NULL, (XtPointer) NULL },
};

/* ARGSUSED */
static void
SFdismissAction(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
  if (event->type == ClientMessage &&
      event->xclient.data.l[0] != SFwmDeleteWindow) return;
  
  SFstatus = SEL_FILE_CANCEL;
}

static char *wmDeleteWindowTranslation = "\
	<Message>WM_PROTOCOLS:	SelFileDismiss()\n\
";

static XtActionsRec actions[] = {
	{"SelFileDismiss",	SFdismissAction},
};


/* position widget under the cursor */
static void
SFpositionWidget(Widget w)
{
  Arg args[3];
  Cardinal num_args;
  Dimension width, height, b_width;
  int x, y, max_x, max_y;
  Window root, child;
  int dummyx, dummyy;
  unsigned int dummymask;
  
  XQueryPointer(XtDisplay(w), XtWindow(w), &root, &child, &x, &y,
                &dummyx, &dummyy, &dummymask);
  num_args = 0;
  XtSetArg(args[num_args], XtNwidth, &width); num_args++;
  XtSetArg(args[num_args], XtNheight, &height); num_args++;
  XtSetArg(args[num_args], XtNborderWidth, &b_width); num_args++;
  XtGetValues(w, args, num_args);
  
  width += 2 * b_width;
  height += 2 * b_width;
  
  x -= ( (Position) width/2 );
  if (x < 0) x = 0;
  if ( x > (max_x = (Position) (XtScreen(w)->width - width)) ) x = max_x;
  
  y -= ( (Position) height/2 );
  if (y < 0) y = 0;
  if ( y > (max_y = (Position) (XtScreen(w)->height - height)) ) y = max_y;
  
  num_args = 0;
  XtSetArg(args[num_args], XtNx, x); num_args++;
  XtSetArg(args[num_args], XtNy, y); num_args++;
  XtSetValues(w, args, num_args);
}

static FILE *
SFopenFile(char *name, char *mode, char *prompt,  char *failed)
{
  Arg args[1];
  FILE *fp;
  
  SFchdir(SFstartDir);
  if ((fp = fopen(name, mode)) == NULL) {
    char *buf;
    buf = XtMalloc((failed ? strlen(failed) : 0) + strlen(strerror(errno)) + 
                   (prompt ? strlen(prompt) : 0) + 2);
    strcpy(buf, failed);
    strcat(buf, strerror(errno));
    strcat(buf, "\n");
    strcat(buf, prompt);
    XtSetArg(args[0], XtNlabel, buf);
    XtSetValues(selFilePrompt, args, ONE);
    XtFree(buf);
    return NULL;
  }
  return fp;
}

static int
SFdeleteEntry(SFDir	*dir, SFEntry	*entry)
{
  register SFEntry	*e;
  register SFEntry	*end;
  int			n;
  int			idx;
  
  idx = entry - dir->entries;
  
  if (idx < dir->beginSelection) {
    dir->beginSelection--;
  }
  if (idx <= dir->endSelection) {
    dir->endSelection--;
  }
  if (dir->beginSelection > dir->endSelection) {
    dir->beginSelection = dir->endSelection = -1;
  }
  
  if (idx < dir->vOrigin) {
    dir->vOrigin--;
  }
  
  XtFree(entry->real);
  
  end = &(dir->entries[dir->nEntries - 1]);
  
  for (e = entry; e < end; e++) {
    *e = *(e + 1);
  }
  
  if (!(--dir->nEntries)) {
    return 1;
  }
  
  n = dir - &(SFdirss[SFdirPtr]);
  if ((n < 0) || (n > 2)) {
    return 1;
  }
  
  XawScrollbarSetThumb(
                       selFileVScrolls[n],
                       (float) (((double) dir->vOrigin) / dir->nEntries),
                       (float) (((double) ((dir->nEntries < SFlistSize) ?
                                           dir->nEntries : SFlistSize)) /
                                dir->nEntries)
                       );
  return 1;
}

static int
SFstatAndCheck(SFDir *dir, SFEntry *entry)
{
  struct stat	statBuf;
  char		save;
  int		last;
  
  /*
   * must be restored before returning
   */
  save = *(dir->path);
  *(dir->path) = 0;
  
  if (!SFchdir(SFcurrentPath)) {
    last = strlen(entry->real) - 1;
    entry->real[last] = 0;
    entry->statDone = 1;
    if (
        (!stat(entry->real, &statBuf))
        
#ifdef S_IFLNK
        
        || (!lstat(entry->real, &statBuf))
        
#endif /* ndef S_IFLNK */
        
        ) {
      if (SFfunc) {
        char *shown;
        
        shown = NULL;
        if (SFfunc(entry->real, &shown, &statBuf)) {
          if (shown) {
            int len;
            
            len = strlen(shown);
            entry->shown = XtMalloc(
                                    (unsigned) (len + 2)
                                    );
            (void) strcpy(entry->shown,
                          shown);
            SFwriteStatChar(
                            entry->shown,
                            len,
                            &statBuf
                            );
            entry->shown[len + 1] = 0;
          }
        } else {
          SFdeleteEntry(dir, entry);
          
          *(dir->path) = save;
          return 1;
        }
      }
      SFwriteStatChar(entry->real, last, &statBuf);
    } else {
      entry->real[last] = ' ';
    }
  }
  
  *(dir->path) = save;
  return 0;
}

static Boolean
SFworkProc(void)
{
  register SFDir		*dir;
  register SFEntry	*entry;
  
  for (dir = &(SFdirss[SFdirEnd - 1]); dir >= SFdirss; dir--) {
    if (!(dir->nEntries)) {
      continue;
    }
    for (
         entry = &(dir->entries[dir->nEntries - 1]);
         entry >= dir->entries;
         entry--
         ) {
      if (!(entry->statDone)) {
        (void) SFstatAndCheck(dir, entry);
        return False;
      }
    }
  }
  
  SFworkProcAdded = 0;
  
  return True;
}

static int
SFclearList(int n, int doScroll)  ;

static int
SFupdatePath(void)
{
  static int	alloc;
  static int	wasTwiddle = 0;
  char		*begin, *end;
  int		i, j;
  int		prevChange;
  int		SFdirPtrSave, SFdirEndSave;
  SFDir		*dir;
  
  if (!SFdirss) {
    SFdirss = (SFDir *) XtMalloc((alloc = 10) * sizeof(SFDir));
    dir = &(SFdirss[0]);
    SFstrdup(&dir->dir, "/");
    (void) SFchdir("/");
    (void) SFgetDir(dir);
    for (j = 1; j < alloc; j++) {
      SFdirss[j].dir = NULL;
    }
    dir->path = SFcurrentPath + 1;
    dir->vOrigin = 0;
    dir->hOrigin = 0;
    dir->changed = 1;
    dir->beginSelection = -1;
    dir->endSelection = -1;
    SFhomeDir.dir = NULL;
    SFrootDir = (SFdirss[0]); /* PK */
  }
  
  SFdirEndSave = SFdirEnd;
  SFdirEnd = 1;
  
  SFdirPtrSave = SFdirPtr;
  SFdirPtr = 0;
  
  begin = NULL;
  
  if (SFcurrentPath[0] == '~') {
    if (!SFtwiddle) {
      SFtwiddle = 1;
      dir = &(SFdirss[0]);
      SFrootDir = *dir;
      if (!SFhomeDir.dir) {
        SFgetHomeDirs();
      }
      *dir = SFhomeDir;
      dir->changed = 1;
    }
    end = SFcurrentPath;
    SFdoNotTouchDirPtr = 1;
    wasTwiddle = 1;
  } else {
    if (SFtwiddle) {
      SFtwiddle = 0;
      dir = &(SFdirss[0]);
      *dir = SFrootDir;
      dir->changed = 1;
    }
    end = SFcurrentPath + 1;
  }
  
  i = 0;
  
  prevChange = 0;
  
  while (*end) {
    while (*end++ == '/') {
      ;
    }
    end--;
    begin = end;
    while ((*end) && (*end++ != '/')) {
      ;
    }
    if ((end - SFcurrentPath <= SFtextPos) && (*(end - 1) == '/')) {
      SFdirPtr = i - 1;
      if (SFdirPtr < 0) {
        SFdirPtr = 0;
      }
    }
    if (*begin) {
      if (*(end - 1) == '/') {
        char save = *end;
        
        if (SFtwiddle) {
          if (SFfindHomeDir(begin, end)) {
            return 1;
          }
        }
        *end = 0;
        i++;
        SFdirEnd++;
        if (i >= alloc) {
          SFdirss = (SFDir *) XtRealloc(
                                       (char *) SFdirss,
                                       (unsigned) ((alloc *= 2) *
                                                   sizeof(SFDir))
                                       );
          for (j = alloc / 2; j < alloc; j++) {
            SFdirss[j].dir = NULL;
          }
        }
        dir = &(SFdirss[i]);
        if (
            (!(dir->dir)) ||
            prevChange ||
            strcmp(dir->dir, begin)
            ) {
          if (dir->dir) {
            SFfree(i);
          }
          prevChange = 1;
          SFstrdup(&dir->dir, begin);
          dir->path = end;
          dir->vOrigin = 0;
          dir->hOrigin = 0;
          dir->changed = 1;
          dir->beginSelection = -1;
          dir->endSelection = -1;
          (void) SFfindFile(dir - 1, begin);
          if (
              SFchdir(SFcurrentPath) ||
              SFgetDir(dir)
              ) {
            SFunreadableDir(dir);
            break;
          }
        }
        *end = save;
        if (!save) {
          SFunselect();
        }
      } else {
        if (SFfindFile(&(SFdirss[SFdirEnd-1]), begin)) {
          return 1;
        }
      }
    } else {
      SFunselect();
    }
  }
  
  if ((end == SFcurrentPath + 1) && (!SFtwiddle)) {
    SFunselect();
  }
  
  for (i = SFdirEnd; i < alloc; i++) {
    if (SFdirss[i].dir) {
      SFfree(i);
    }
  }
  
  if (SFdoNotTouchDirPtr) {
    if (wasTwiddle) {
      wasTwiddle = 0;
      SFdirPtr = SFdirEnd - 2;
      if (SFdirPtr < 0) {
        SFdirPtr = 0;
      }
    } else {
      SFdirPtr = SFdirPtrSave;
    }
    SFdoNotTouchDirPtr = 0;
  }
  
  if ((SFdirPtr != SFdirPtrSave) || (SFdirEnd != SFdirEndSave)) {
    XawScrollbarSetThumb(
                         selFileHScroll,
                         (float) (((double) SFdirPtr) / SFdirEnd),
                         (float) (((double) ((SFdirEnd < 3) ? SFdirEnd : 3)) /
                                  SFdirEnd)
                         );
  }
  
  if (SFdirPtr != SFdirPtrSave) {
    SFdrawLists(SF_DO_SCROLL);
  } else {
    for (i = 0; i < 3; i++) {
      if (SFdirPtr + i < SFdirEnd) {
        if (SFdirss[SFdirPtr + i].changed) {
          SFdirss[SFdirPtr + i].changed = 0;
          SFdrawList(i, SF_DO_SCROLL);
        }
      } else {
        SFclearList(i, SF_DO_SCROLL);
      }
    }
  }
  return 1;
}

static int
SFtextChanged(void)
{
  
  if ((SFtextBuffer[0] == '/') || (SFtextBuffer[0] == '~')) {
    (void) strcpy(SFcurrentPath, SFtextBuffer);
    
    SFtextPos = XawTextGetInsertionPoint(selFileField);
  } else {
    (void) strcat(strcpy(SFcurrentPath, SFstartDir), SFtextBuffer);
    
    SFtextPos = XawTextGetInsertionPoint(selFileField) +
      strlen(SFstartDir);
  }
  
  if (!SFworkProcAdded) {
    (void) XtAppAddWorkProc(SFapp, (XtWorkProc) SFworkProc, NULL);
    SFworkProcAdded = 1;
  }
  
  SFupdatePath();
  return 1;
}

static char *
SFgetText(void)
{
  return strcpy(XtMalloc((unsigned) (strlen(SFtextBuffer) + 1)),
		SFtextBuffer);
}

static int
SFprepareToReturn(void)
{
  SFstatus = SEL_FILE_NULL;
  XtRemoveGrab(selFile);
  XtUnmapWidget(selFile);
  XtRemoveTimeOut(SFdirModTimerId);
  if (SFchdir(SFstartDir)) {
    XtAppError(
               SFapp,
               "XsraSelFile: can't return to current directory"
               );
  }
  return 1;
}

static void
SFdirModTimer(XtPointer cl, XtIntervalId    *id)
{
  static int	n = -1;
  static int	f = 0;
  char		save;
  SFDir		*dir;
  
  if ((!SFtwiddle) && (SFdirPtr < SFdirEnd)) {
    n++;
    if ((n > 2) || (SFdirPtr + n >= SFdirEnd)) {
      n = 0;
      f++;
      if ((f > 2) || (SFdirPtr + f >= SFdirEnd)) {
        f = 0;
      }
    }
    dir = &(SFdirss[SFdirPtr + n]);
    save = *(dir->path);
    *(dir->path) = 0;
    if (SFchdir(SFcurrentPath)) {
      *(dir->path) = save;
      
      /*
       * force a re-read
       */
      *(dir->dir) = 0;
      
      SFupdatePath();
    } else {
      *(dir->path) = save;
      if (
          SFcheckDir(n, dir) ||
          ((f == n) && SFcheckFiles(dir))
          ) {
        SFdrawList(n, SF_DO_SCROLL);
      }
    }
  }
  
  SFdirModTimerId = XtAppAddTimeOut(SFapp, (unsigned long) 1000,
                                    SFdirModTimer, (XtPointer) NULL);
}


/*
 * Copyright 1989 Software Research Associates, Inc., Tokyo, Japan
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be used
 * in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  Software Research Associates
 * makes no representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 *
 * SOFTWARE RESEARCH ASSOCIATES DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL SOFTWARE RESEARCH ASSOCIATES BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik M. van der Poel
 *         Software Research Associates, Inc., Tokyo, Japan
 *         erik@sra.co.jp
 */

#include "Econfig.h"
#include "Ecfg.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <pwd.h>
#ifdef SEL_FILE_IGNORE_CASE
#include <ctype.h>
#endif /* def SEL_FILE_IGNORE_CASE */

#include <X11/Xos.h>
#include <pwd.h>
#include "Eselfile.h"

#ifdef HAVE_GETUID
extern uid_t getuid();
extern void qsort();
#endif 



#ifdef SEL_FILE_IGNORE_CASE
static int
SFstrncmp(register char	*p, register char *q, register int n)
{
  register char	c1, c2;
  char		*psave, *qsave;
  int		nsave;
  
  psave = p;
  qsave = q;
  nsave = n;
  
  c1 = *p++;
  if (islower(c1)) {
    c1 = toupper(c1);
  }
  c2 = *q++;
  if (islower(c2)) {
    c2 = toupper(c2);
  }
  
  while ((--n >= 0) && (c1 == c2)) {
    if (!c1) {
      return strncmp(psave, qsave, nsave);
    }
    c1 = *p++;
    if (islower(c1)) {
      c1 = toupper(c1);
    }
    c2 = *q++;
    if (islower(c2)) {
      c2 = toupper(c2);
    }
  }
  
  if (n < 0) {
    return strncmp(psave, qsave, nsave);
  }
  
  return c1 - c2;
}
#endif /* def SEL_FILE_IGNORE_CASE */

static int
SFsetText(char *path)
{
  XawTextBlock	text;
  
  text.firstPos = 0;
  text.length = strlen(path);
  text.ptr = path;
  text.format = FMT8BIT;
  
  XawTextReplace(selFileField, 0, strlen(SFtextBuffer), &text);
  XawTextSetInsertionPoint(selFileField, strlen(SFtextBuffer));
  return 1;
}

static int
SFreplaceText(SFDir	*dir, char	*str)
{
  int	len;
  
  *(dir->path) = 0;
  len = strlen(str);
  if (str[len - 1] == '/') {
    (void) strcat(SFcurrentPath, str);
  } else {
    (void) strncat(SFcurrentPath, str, len - 1);
  }
  if (strncmp(SFcurrentPath, SFstartDir, strlen(SFstartDir))) {
    SFsetText(SFcurrentPath);
  } else {
    SFsetText(&(SFcurrentPath[strlen(SFstartDir)]));
  }
  
  SFtextChanged();
  return 1;
}

static void
SFexpand(char	*str)
{
  int	len;
  int	cmp;
  char	*name, *growing;
  SFDir	*dir;
  SFEntry	*entry, *max;
  
  len = strlen(str);
  
  dir = &(SFdirss[SFdirEnd - 1]);
  
  if (dir->beginSelection == -1) {
    SFstrdup(&str, str);
    SFreplaceText(dir, str);
    XtFree(str);
    return;
  } else if (dir->beginSelection == dir->endSelection) {
    SFreplaceText(dir, dir->entries[dir->beginSelection].shown);
    return;
  }
  
  max = &(dir->entries[dir->endSelection + 1]);
  
  name = dir->entries[dir->beginSelection].shown;
  SFstrdup(&growing, name);
  
  cmp = 0;
  while (!cmp) {
    entry = &(dir->entries[dir->beginSelection]);
    while (entry < max) {
      if ((cmp = strncmp(growing, entry->shown, len))) {
        break;
      }
      entry++;
    }
    len++;
  }
  
  /*
   * SFreplaceText() expects filename
   */
  growing[len - 2] = ' ';
  
  growing[len - 1] = 0;
  SFreplaceText(dir, growing);
  XtFree(growing);
}

static int
SFfindFile(SFDir		*dir, register char	*str)
{
  register int	i, last, max;
  register char	*name, save;
  SFEntry		*entries;
  int		len;
  int		begin, end;
  int		result;
  
  len = strlen(str);
  
  if (str[len - 1] == ' ') {
    SFexpand(str);
    return 1;
  } else if (str[len - 1] == '/') {
    len--;
  }
  
  max = dir->nEntries;
  
  entries = dir->entries;
  
  i = 0;
  while (i < max) {
    name = entries[i].shown;
    last = strlen(name) - 1;
    save = name[last];
    name[last] = 0;
    
#ifdef SEL_FILE_IGNORE_CASE
    result = SFstrncmp(str, name, len);
#else /* def SEL_FILE_IGNORE_CASE */
    result = strncmp(str, name, len);
#endif /* def SEL_FILE_IGNORE_CASE */
    
    name[last] = save;
    if (result <= 0) {
      break;
    }
    i++;
  }
  begin = i;
  while (i < max) {
    name = entries[i].shown;
    last = strlen(name) - 1;
    save = name[last];
    name[last] = 0;
    
#ifdef SEL_FILE_IGNORE_CASE
    result = SFstrncmp(str, name, len);
#else /* def SEL_FILE_IGNORE_CASE */
    result = strncmp(str, name, len);
#endif /* def SEL_FILE_IGNORE_CASE */
    
    name[last] = save;
    if (result) {
      break;
    }
    i++;
  }
  end = i;
  
  if (begin != end) {
    if (
        (dir->beginSelection != begin) ||
        (dir->endSelection != end - 1)
        ) {
      dir->changed = 1;
      dir->beginSelection = begin;
      if (str[strlen(str) - 1] == '/') {
        dir->endSelection = begin;
      } else {
        dir->endSelection = end - 1;
      }
    }
  } else {
    if (dir->beginSelection != -1) {
      dir->changed = 1;
      dir->beginSelection = -1;
      dir->endSelection = -1;
    }
  }
  
  if (
      SFdoNotTouchVorigin ||
      ((begin > dir->vOrigin) && (end < dir->vOrigin + SFlistSize))
      ) {
    SFdoNotTouchVorigin = 0;
    return 0;
  }
  
  i = begin - 1;
  if (i > max - SFlistSize) {
    i = max - SFlistSize;
  }
  if (i < 0) {
    i = 0;
  }
  
  if (dir->vOrigin != i) {
    dir->vOrigin = i;
    dir->changed = 1;
  }
  
  return 0;
}

static int
SFunselect(void)
{
  SFDir	*dir;
  
  dir = &(SFdirss[SFdirEnd - 1]);
  if (dir->beginSelection != -1) {
    dir->changed = 1;
  }
  dir->beginSelection = -1;
  dir->endSelection = -1;
  return 1;
}

static int
SFcompareLogins(SFLogin	*p, SFLogin	*q)
{
  return strcmp(p->name, q->name);
}

static int
SFgetHomeDirs(void)
{
  struct passwd	*pw;
  int		alloc;
  int		i;
  SFEntry		*entries = NULL;
  int		len;
  int		maxChars;
  
  {
    alloc = 1;
    i = 1;
    entries = (SFEntry *) XtMalloc(sizeof(SFEntry));
    SFlogins = (SFLogin *) XtMalloc(sizeof(SFLogin));
    entries[0].real = XtMalloc(3);
    (void) strcpy(entries[0].real, "~");
    entries[0].shown = entries[0].real;
    entries[0].statDone = 1;
    SFlogins[0].name = "";
    pw = getpwuid((int) getuid());
    SFstrdup(&SFlogins[0].dir, pw ? pw->pw_dir : "/");
    maxChars = 0;
  }
  
  (void) setpwent();
  
  while ((pw = getpwent()) && (*(pw->pw_name))) {
    if (i >= alloc) {
      alloc *= 2;
      entries = (SFEntry *) XtRealloc(
                                      (char *) entries,
                                      (unsigned) (alloc * sizeof(SFEntry))
                                      );
      SFlogins = (SFLogin *) XtRealloc(
                                       (char *) SFlogins,
                                       (unsigned) (alloc * sizeof(SFLogin))
                                       );
    }
    len = strlen(pw->pw_name);
    entries[i].real = XtMalloc((unsigned) (len + 3));
    (void) strcat(strcpy(entries[i].real, "~"),
                  pw->pw_name);
    entries[i].shown = entries[i].real;
    entries[i].statDone = 1;
    if (len > maxChars) {
      maxChars = len;
    }
    SFstrdup(&SFlogins[i].name, pw->pw_name);
    SFstrdup(&SFlogins[i].dir, pw->pw_dir);
    i++;
  }
  
  SFhomeDir.dir			= XtMalloc(1)	;
  SFhomeDir.dir[0]		= 0		;
  SFhomeDir.path			= SFcurrentPath	;
  SFhomeDir.entries		= entries	;
  SFhomeDir.nEntries		= i		;
  SFhomeDir.vOrigin		= 0		;	/* :-) */
  SFhomeDir.nChars		= maxChars + 2	;
  SFhomeDir.hOrigin		= 0		;
  SFhomeDir.changed		= 1		;
  SFhomeDir.beginSelection	= -1		;
  SFhomeDir.endSelection		= -1		;
  
#define CST_ (int (*) (const void *, const void *))
#if defined(SVR4) || defined(SYSV) || defined(USG)
  qsort((char *) entries, (unsigned)i, sizeof(SFEntry), CST_ SFcompareEntries);
  qsort((char *) SFlogins, (unsigned)i, sizeof(SFLogin), CST_ SFcompareLogins);
#else /* defined(SVR4) || defined(SYSV) || defined(USG) */
  qsort((char *) entries, i, sizeof(SFEntry), CST_ SFcompareEntries);
  qsort((char *) SFlogins, i, sizeof(SFLogin), CST_ SFcompareLogins);
#endif /* defined(SVR4) || defined(SYSV) || defined(USG) */
  
  for (i--; i >= 0; i--) {
    (void) strcat(entries[i].real, "/");
  }
  return 1;
}

static int
SFfindHomeDir(char	*begin, char	*end)
{
  char	save;
  char	*theRest;
  int	i;
  
  save = *end;
  *end = 0;
  
  for (i = SFhomeDir.nEntries - 1; i >= 0; i--) {
    if (!strcmp(SFhomeDir.entries[i].real, begin)) {
      *end = save;
      SFstrdup(&theRest, end);
      (void) strcat(strcat(strcpy(SFcurrentPath,
                                  SFlogins[i].dir), "/"), theRest);
      XtFree(theRest);
      SFsetText(SFcurrentPath);
      SFtextChanged();
      return 1;
    }
  }
  
  *end = save;
  
  return 0;
}



/* ARGSUSED */
static void
SFbuttonPressList(Widget w, int n, XButtonPressedEvent	*event)
{
  SFbuttonPressed = 1;
}
static void
SFmotionList(Widget w, register int n, register XButtonPressedEvent *event);

/* ARGSUSED */
static void
SFbuttonReleaseList(Widget w, int n, XButtonReleasedEvent *event)
{
  SFDir	*dir;
  
  SFbuttonPressed = 0;
  
  if (SFcurrentInvert[n] != -1) {
    if (n < 2) {
      SFdoNotTouchDirPtr = 1;
    }
    SFdoNotTouchVorigin = 1;
    dir = &(SFdirss[SFdirPtr + n]);
    SFreplaceText(
                  dir,
                  dir->entries[dir->vOrigin + SFcurrentInvert[n]].shown
                  );
    SFmotionList(w, n, event);
  }
}

static int
SFcheckDir(int n, SFDir		*dir)
{
  struct stat	statBuf;
  int		i;
  
  if (
      (!stat(".", &statBuf)) &&
      (statBuf.st_mtime != dir->mtime)
      ) {
    
    /*
     * If the pointer is currently in the window that we are about
     * to update, we must warp it to prevent the user from
     * accidentally selecting the wrong file.
     */
    if (SFcurrentInvert[n] != -1) {
      XWarpPointer(
                   SFdisplay,
                   None,
                   XtWindow(selFileLists[n]),
                   0,
                   0,
                   0,
                   0,
                   0,
                   0
                   );
    }
    
    for (i = dir->nEntries - 1; i >= 0; i--) {
      if (dir->entries[i].shown != dir->entries[i].real) {
        XtFree(dir->entries[i].shown);
      }
      XtFree(dir->entries[i].real);
    }
    XtFree((char *) dir->entries);
    if (SFgetDir(dir)) {
      SFunreadableDir(dir);
    }
    if (dir->vOrigin > dir->nEntries - SFlistSize) {
      dir->vOrigin = dir->nEntries - SFlistSize;
    }
    if (dir->vOrigin < 0) {
      dir->vOrigin = 0;
    }
    if (dir->hOrigin > dir->nChars - SFcharsPerEntry) {
      dir->hOrigin = dir->nChars - SFcharsPerEntry;
    }
    if (dir->hOrigin < 0) {
      dir->hOrigin = 0;
    }
    dir->beginSelection = -1;
    dir->endSelection = -1;
    SFdoNotTouchVorigin = 1;
    if ((dir + 1)->dir) {
      (void) SFfindFile(dir, (dir + 1)->dir);
    } else {
      (void) SFfindFile(dir, dir->path);
    }
    
    if (!SFworkProcAdded) {
      (void) XtAppAddWorkProc(SFapp, (XtWorkProc) SFworkProc, NULL);
      SFworkProcAdded = 1;
    }
    
    return 1;
  }
  
  return 0;
}

/* Return a single character describing what kind of file STATBUF is.  */

static char
SFstatChar1 (struct stat *statBuf)
{
  if (S_ISDIR (statBuf->st_mode)) {
    return '/';
  } else if (S_ISREG (statBuf->st_mode)) {
    return S_ISXXX (statBuf->st_mode) ? '*' : ' ';
#ifdef S_ISSOCK
  } else if (S_ISSOCK (statBuf->st_mode)) {
    return '=';
#endif /* S_ISSOCK */
  } else {
    return ' ';
  }
}

static int
SFcheckFiles(SFDir	*dir)
{
  int		from, to;
  int		result;
  char		old, new;
  int		i;
  char		*str;
  int		last;
  struct stat	statBuf;
  
  result = 0;
  
  from = dir->vOrigin;
  to = dir->vOrigin + SFlistSize;
  if (to > dir->nEntries) {
    to = dir->nEntries;
  }
  
  for (i = from; i < to; i++) {
    str = dir->entries[i].real;
    last = strlen(str) - 1;
    old = str[last];
    str[last] = 0;
    if (stat(str, &statBuf)) {
      new = ' ';
    } else {
      new = SFstatChar1(&statBuf);
    }
    str[last] = new;
    if (new != old) {
      result = 1;
    }
  }
  
  return result;
}





/* drawl.c */
/*
 * Copyright 1989 Software Research Associates, Inc., Tokyo, Japan
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Software Research Associates not be used
 * in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  Software Research Associates
 * makes no representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 *
 * SOFTWARE RESEARCH ASSOCIATES DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL SOFTWARE RESEARCH ASSOCIATES BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik M. van der Poel
 *         Software Research Associates, Inc., Tokyo, Japan
 *         erik@sra.co.jp
 */

#include "Econfig.h"
#include "Ecfg.h"
#include <stdio.h>
#include "Eselfile.h"
#include <X11/StringDefs.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Cardinals.h>


static int
SFinitFont(void)
{
  TextData	*data;
  
  data = XtNew(TextData);
  
  XtGetApplicationResources(selFileForm, (XtPointer) data, textResources,
                            XtNumber(textResources), (Arg *) NULL, 0);
  
  SFfont = XLoadQueryFont(SFdisplay, data->fontname);
  if (!SFfont) {
    SFfont = XLoadQueryFont(SFdisplay, SF_DEFAULT_FONT);
    if (!SFfont) {
      char	sbuf[256];
      
      (void) sprintf(sbuf, "XsraSelFile: can't get font %s",
                     SF_DEFAULT_FONT);
      
      XtAppError(SFapp, sbuf);
    }
  }
  
  SFcharWidth = (SFfont->max_bounds.width + SFfont->min_bounds.width) / 2;
  SFcharAscent = SFfont->max_bounds.ascent;
  SFcharHeight = SFcharAscent + SFfont->max_bounds.descent;
  return 1;
}

static int
SFcreateGC(void)
{
  XGCValues	gcValues;
  XRectangle	rectangles[1];
  
  gcValues.foreground = SFfore;
  
  SFlineGC = XtGetGC(
                     selFileLists[0],
                     (XtGCMask)
                     GCForeground		|
                     0,
                     &gcValues
                     );
  
  SFscrollGC = XtGetGC(
                       selFileLists[0],
                       (XtGCMask)
                       0,
                       &gcValues
                       );
  
  gcValues.function = GXinvert;
  gcValues.plane_mask = (SFfore ^ SFback);
  
  SFinvertGC = XtGetGC(
                       selFileLists[0],
                       (XtGCMask)
                       GCFunction		|
                       GCPlaneMask		|
                       0,
                       &gcValues
                       );
  
  gcValues.foreground = SFfore;
  gcValues.background = SFback;
  gcValues.font = SFfont->fid;
  
  SFtextGC = XCreateGC(
                       SFdisplay,
                       XtWindow(selFileLists[0]),
                       (unsigned long)
                       GCForeground		|
                       GCBackground		|
                       GCFont			|
                       0,
                       &gcValues
                       );
  
  rectangles[0].x = SFlineToTextH + SFbesideText;
  rectangles[0].y = 0;
  rectangles[0].width = SFcharsPerEntry * SFcharWidth;
  rectangles[0].height = SFupperY + 1;
  
  XSetClipRectangles(
                     SFdisplay,
                     SFtextGC,
                     0,
                     0,
                     rectangles,
                     1,
                     Unsorted
                     );
  return 1;
}

static int
SFclearList(int n, int doScroll)
{
  SFDir	*dir;
  
  SFcurrentInvert[n] = -1;
  
  XClearWindow(SFdisplay, XtWindow(selFileLists[n]));
  
  XDrawSegments(SFdisplay, XtWindow(selFileLists[n]), SFlineGC, SFsegs, 2);
  
  if (doScroll) {
    dir = &(SFdirss[SFdirPtr + n]);
    
    if ((SFdirPtr + n < SFdirEnd) && dir->nEntries && dir->nChars) {
      XawScrollbarSetThumb(
                           selFileVScrolls[n],
                           (float) (((double) dir->vOrigin) /
                                    dir->nEntries),
                           (float) (((double) ((dir->nEntries < SFlistSize)
                                               ? dir->nEntries : SFlistSize)) /
                                    dir->nEntries)
                           );
      
      XawScrollbarSetThumb(
                           selFileHScrolls[n],
                           (float) (((double) dir->hOrigin) / dir->nChars),
                           (float) (((double) ((dir->nChars <
                                                SFcharsPerEntry) ? dir->nChars :
                                               SFcharsPerEntry)) / dir->nChars)
                           );
    } else {
      XawScrollbarSetThumb(selFileVScrolls[n], (float) 0.0,
                           (float) 1.0);
      XawScrollbarSetThumb(selFileHScrolls[n], (float) 0.0,
                           (float) 1.0);
    }
  }
  return 1;
}


static int
SFwriteStatChar(char *name, int last, struct stat *statBuf)
{
  name[last] = SFstatChar1(statBuf);
  return 1;
}



static int
SFdrawStrings(register Window w, register SFDir *dir,
              register int from, register int to)
{
  register int		i;
  register SFEntry	*entry;
  int			x;
  
  x = SFtextX - dir->hOrigin * SFcharWidth;
  
  if (dir->vOrigin + to >= dir->nEntries) {
    to = dir->nEntries - dir->vOrigin - 1;
  }
  for (i = from; i <= to; i++) {
    entry = &(dir->entries[dir->vOrigin + i]);
    if (!(entry->statDone)) {
      if (SFstatAndCheck(dir, entry)) {
        if (dir->vOrigin + to >= dir->nEntries) {
          to = dir->nEntries - dir->vOrigin - 1;
        }
        i--;
        continue;
      }
    }
    XDrawImageString(
                     SFdisplay,
                     w,
                     SFtextGC,
                     x,
                     SFtextYoffset + i * SFentryHeight,
                     entry->shown,
                     strlen(entry->shown)
                     );
    if (dir->vOrigin + i == dir->beginSelection) {
      XDrawLine(
                SFdisplay,
                w,
                SFlineGC,
                SFlineToTextH + 1,
                SFlowerY + i * SFentryHeight,
                SFlineToTextH + SFentryWidth - 2,
                SFlowerY + i * SFentryHeight
                );
    }
    if (
        (dir->vOrigin + i >= dir->beginSelection) &&
        (dir->vOrigin + i <= dir->endSelection)
        ) {
      SFcompletionSegs[0].y1 = SFcompletionSegs[1].y1 =
        SFlowerY + i * SFentryHeight;
      SFcompletionSegs[0].y2 = SFcompletionSegs[1].y2 =
        SFlowerY + (i + 1) * SFentryHeight - 1;
      XDrawSegments(
                    SFdisplay,
                    w,
                    SFlineGC,
                    SFcompletionSegs,
                    2
                    );
    }
    if (dir->vOrigin + i == dir->endSelection) {
      XDrawLine(
                SFdisplay,
                w,
                SFlineGC,
                SFlineToTextH + 1,
                SFlowerY + (i + 1) * SFentryHeight - 1,
                SFlineToTextH + SFentryWidth - 2,
                SFlowerY + (i + 1) * SFentryHeight - 1
                );
    }
  }
  return 1;
}

static int
SFdrawList(int n, int doScroll)
{
  SFDir	*dir;
  Window	w;
  
  SFclearList(n, doScroll);
  
  if (SFdirPtr + n < SFdirEnd) {
    dir = &(SFdirss[SFdirPtr + n]);
    w = XtWindow(selFileLists[n]);
    XDrawImageString(
                     SFdisplay,
                     w,
                     SFtextGC,
                     SFtextX - dir->hOrigin * SFcharWidth,
                     SFlineToTextV + SFaboveAndBelowText + SFcharAscent,
                     dir->dir,
                     strlen(dir->dir)
                     );
    SFdrawStrings(w, dir, 0, SFlistSize - 1);
  }
  return 1;
}

static int
SFdrawLists(int doScroll)
{
  int	i;
  
  for (i = 0; i < 3; i++) {
    SFdrawList(i, doScroll);
  }
  return 1;
}

static int
SFinvertEntry(int n)
{
  XFillRectangle(
                 SFdisplay,
                 XtWindow(selFileLists[n]),
                 SFinvertGC,
                 SFlineToTextH,
                 SFcurrentInvert[n] * SFentryHeight + SFlowerY,
                 SFentryWidth,
                 SFentryHeight
                 );
  return 1;
}

static unsigned long
SFscrollTimerInterval(void)
{
  static int	maxVal = 200;
  static int	varyDist = 50;
  static int	minDist = 50;
  int		t;
  int		dist;
  
  if (SFcurrentListY < SFlowerY) {
    dist = SFlowerY - SFcurrentListY;
  } else if (SFcurrentListY > SFupperY) {
    dist = SFcurrentListY - SFupperY;
  } else {
    return (unsigned long) 1;
  }
  
  t = maxVal - ((maxVal / varyDist) * (dist - minDist));
  
  if (t < 1) {
    t = 1;
  }
  
  if (t > maxVal) {
    t = maxVal;
  }
  
  return (unsigned long) t;
}

static void
SFvSliderMovedCallback(Widget w, int n, int new);

static void
SFscrollTimer(XtPointer p, XtIntervalId *id)
{
  SFDir	*dir;
  int	save;
  int     n;
  
  n = (int) p;
  
  dir = &(SFdirss[SFdirPtr + n]);
  save = dir->vOrigin;
  
  if (SFcurrentListY < SFlowerY) {
    if (dir->vOrigin > 0) {
      SFvSliderMovedCallback(selFileVScrolls[n], n,
                             dir->vOrigin - 1);
    }
  } else if (SFcurrentListY > SFupperY) {
    if (dir->vOrigin < dir->nEntries - SFlistSize) {
      SFvSliderMovedCallback(selFileVScrolls[n], n,
                             dir->vOrigin + 1);
    }
  }
  
  if (dir->vOrigin != save) {
    if (dir->nEntries) {
      XawScrollbarSetThumb(
                           selFileVScrolls[n],
                           (float) (((double) dir->vOrigin) / dir->nEntries),
                           (float) (((double) ((dir->nEntries < SFlistSize) ?
                                               dir->nEntries : SFlistSize)) / dir->nEntries)
                           );
    }
  }
  
  if (SFbuttonPressed) {
    SFscrollTimerId = XtAppAddTimeOut(SFapp,
                                      SFscrollTimerInterval(), SFscrollTimer, (XtPointer) n);
  }
}

static int
SFnewInvertEntry(register int n, register XButtonReleasedEvent *event)
{
  register int	x, y;
  register int	new;
  static int	SFscrollTimerAdded = 0;
  
  x = event->x;
  y = event->y;
  
  if (SFdirPtr + n >= SFdirEnd) {
    return -1;
  } else if (
             (x >= 0)	&& (x <= SFupperX) &&
             (y >= SFlowerY)	&& (y <= SFupperY)
             ) {
    register SFDir *dir = &(SFdirss[SFdirPtr + n]);
    
    if (SFscrollTimerAdded) {
      SFscrollTimerAdded = 0;
      XtRemoveTimeOut(SFscrollTimerId);
    }
    
    new = (y - SFlowerY) / SFentryHeight;
    if (dir->vOrigin + new >= dir->nEntries) {
      return -1;
    }
    return new;
  } else {
    if (SFbuttonPressed) {
      SFcurrentListY = y;
      if (!SFscrollTimerAdded) {
        SFscrollTimerAdded = 1;
        SFscrollTimerId = XtAppAddTimeOut(SFapp,
                                          SFscrollTimerInterval(), SFscrollTimer,
                                          (XtPointer) n);
      }
    }
    
    return -1;
  }
}

/* ARGSUSED */
static void
SFenterList(Widget w, register int n, register XEnterWindowEvent *event)
{
  register int	new;
  
  /* sanity */
  if (SFcurrentInvert[n] != -1) {
    SFinvertEntry(n);
    SFcurrentInvert[n] = -1;
  }
  
  new = SFnewInvertEntry(n, (XButtonReleasedEvent *) event);
  if (new != -1) {
    SFcurrentInvert[n] = new;
    SFinvertEntry(n);
  }
}

/* ARGSUSED */
static void
SFleaveList(Widget w, register int n, XEvent *event)
{
  if (SFcurrentInvert[n] != -1) {
    SFinvertEntry(n);
    SFcurrentInvert[n] = -1;
  }
}

/* ARGSUSED */
static void
SFmotionList(Widget w, register int n, register XButtonReleasedEvent *event)
{
  register int	new;
  
  new = SFnewInvertEntry(n, event);
  
  if (new != SFcurrentInvert[n]) {
    if (SFcurrentInvert[n] != -1) {
      SFinvertEntry(n);
    }
    SFcurrentInvert[n] = new;
    if (new != -1) {
      SFinvertEntry(n);
    }
  }
}

/* ARGSUSED */
static void
SFvFloatSliderMovedCallback(Widget w, int n, float	*fnew)
{
  int	new;
  
  new = (*fnew) * SFdirss[SFdirPtr + n].nEntries;
  
  SFvSliderMovedCallback(w, n, new);
}

/* ARGSUSED */
static void
SFvSliderMovedCallback(Widget w, int n, int new)
{
  int		old;
  register Window	win;
  SFDir		*dir;
  
  dir = &(SFdirss[SFdirPtr + n]);
  
  old = dir->vOrigin;
  dir->vOrigin = new;
  
  if (old == new) {
    return;
  }
  
  win = XtWindow(selFileLists[n]);
  
  if (ABS(new - old) < SFlistSize) {
    if (new > old) {
      XCopyArea(
                SFdisplay,
                win,
                win,
                SFscrollGC,
                SFlineToTextH,
                SFlowerY + (new - old) * SFentryHeight,
                SFentryWidth + SFlineToTextH,
                (SFlistSize - (new - old)) * SFentryHeight,
                SFlineToTextH,
                SFlowerY
                );
      XClearArea(
                 SFdisplay,
                 win,
                 SFlineToTextH,
                 SFlowerY + (SFlistSize - (new - old)) *
                 SFentryHeight,
                 SFentryWidth + SFlineToTextH,
                 (new - old) * SFentryHeight,
                 False
                 );
      SFdrawStrings(win, dir, SFlistSize - (new - old),
                    SFlistSize - 1);
    } else {
      XCopyArea(
                SFdisplay,
                win,
                win,
                SFscrollGC,
                SFlineToTextH,
                SFlowerY,
                SFentryWidth + SFlineToTextH,
                (SFlistSize - (old - new)) * SFentryHeight,
                SFlineToTextH,
                SFlowerY + (old - new) * SFentryHeight
                );
      XClearArea(
                 SFdisplay,
                 win,
                 SFlineToTextH,
                 SFlowerY,
                 SFentryWidth + SFlineToTextH,
                 (old - new) * SFentryHeight,
                 False
                 );
      SFdrawStrings(win, dir, 0, old - new);
    }
  } else {
    XClearArea(
               SFdisplay,
               win,
               SFlineToTextH,
               SFlowerY,
               SFentryWidth + SFlineToTextH,
               SFlistSize * SFentryHeight,
               False
               );
    SFdrawStrings(win, dir, 0, SFlistSize - 1);
  }
}

/* ARGSUSED */
static void
SFvAreaSelectedCallback(Widget w, int n, int pnew)
{
  SFDir	*dir;
  int	new;
  
  dir = &(SFdirss[SFdirPtr + n]);
  
  new = dir->vOrigin +
    (((double) pnew) / SFvScrollHeight) * dir->nEntries;
  
  if (new > dir->nEntries - SFlistSize) {
    new = dir->nEntries - SFlistSize;
  }
  
  if (new < 0) {
    new = 0;
  }
  
  if (dir->nEntries) {
    float	f;
    
    f = ((double) new) / dir->nEntries;
    
    XawScrollbarSetThumb(
                         w,
                         f,
                         (float) (((double) ((dir->nEntries < SFlistSize) ?
                                             dir->nEntries : SFlistSize)) /
                                  dir->nEntries)
                         );
  }
  
  SFvSliderMovedCallback(w, n, new);
}

/* ARGSUSED */
static void
SFhSliderMovedCallback(Widget w, int n, float *new)
{
  SFDir	*dir;
  int	save;
  
  dir = &(SFdirss[SFdirPtr + n]);
  save = dir->hOrigin;
  dir->hOrigin = (*new) * dir->nChars;
  if (dir->hOrigin == save) {
    return;
  }
  
  SFdrawList(n, SF_DO_NOT_SCROLL);
}

/* ARGSUSED */
static void
SFhAreaSelectedCallback(Widget w, int n, int pnew)
{
  SFDir	*dir;
  int	new;
  
  dir = &(SFdirss[SFdirPtr + n]);
  
  new = dir->hOrigin +
    (((double) pnew) / SFhScrollWidth) * dir->nChars;
  
  if (new > dir->nChars - SFcharsPerEntry) {
    new = dir->nChars - SFcharsPerEntry;
  }
  
  if (new < 0) {
    new = 0;
  }
  
  if (dir->nChars) {
    float	f;
    
    f = ((double) new) / dir->nChars;
    
    XawScrollbarSetThumb(
                         w,
                         f,
                         (float) (((double) ((dir->nChars < SFcharsPerEntry) ?
                                             dir->nChars : SFcharsPerEntry)) / dir->nChars)
                         );
    
    SFhSliderMovedCallback(w, n, &f);
  }
}

/* ARGSUSED */
static void
SFpathSliderMovedCallback(Widget w, XtPointer client_data, float *new)
{
  SFDir		*dir;
  int		n;
  XawTextPosition	pos;
  int	SFdirPtrSave;
  
  SFdirPtrSave = SFdirPtr;
  SFdirPtr = (*new) * SFdirEnd;
  if (SFdirPtr == SFdirPtrSave) {
    return;
  }
  
  SFdrawLists(SF_DO_SCROLL);
  
  n = 2;
  while (SFdirPtr + n >= SFdirEnd) {
    n--;
  }
  
  dir = &(SFdirss[SFdirPtr + n]);
  
  pos = dir->path - SFcurrentPath;
  
  if (!strncmp(SFcurrentPath, SFstartDir, strlen(SFstartDir))) {
    pos -= strlen(SFstartDir);
    if (pos < 0) {
      pos = 0;
    }
  }
  
  XawTextSetInsertionPoint(selFileField, pos);
}

/* ARGSUSED */

static void
SFpathAreaSelectedCallback(Widget w, XtPointer client_data, int pnew)
{
  int	new;
  float	f;
  
  new = SFdirPtr + (((double) pnew) / SFpathScrollWidth) * SFdirEnd;
  
  if (new > SFdirEnd - 3) {
    new = SFdirEnd - 3;
  }
  
  if (new < 0) {
    new = 0;
  }
  
  f = ((double) new) / SFdirEnd;
  
  XawScrollbarSetThumb(
                       w,
                       f,
                       (float) (((double) ((SFdirEnd < 3) ? SFdirEnd : 3)) /
                                SFdirEnd)
                       );
  
  SFpathSliderMovedCallback(w, (XtPointer) NULL, &f);
}



static void
SFcreateWidgets(Widget toplevel, char *prompt, char *ok, char *cancel)
{
  Cardinal	i, n;
  int		listWidth, listHeight;
  int		listSpacing = 10;
  int		scrollThickness = 15;
  int		hScrollX, hScrollY;
  int		vScrollX, vScrollY;
  Cursor
    xtermCursor,
    sbRightArrowCursor,
    dotCursor;
  Arg		arglist[20];
  
  i = 0;
  XtSetArg(arglist[i], XtNtransientFor, toplevel);		i++;
  
  selFile = XtCreatePopupShell("selFile", 
                               transientShellWidgetClass, ESITopLevelWidget(),
                               arglist, i);

  /* Add WM_DELETE_WINDOW protocol */
  XtAppAddActions(XtWidgetToApplicationContext(selFile),
                  actions, XtNumber(actions));
  XtOverrideTranslations(selFile,
                         XtParseTranslationTable(wmDeleteWindowTranslation));
  
  i = 0;
  XtSetArg(arglist[i], XtNdefaultDistance, 30);			i++;
  selFileForm = XtCreateManagedWidget("selFileForm",
                                      formWidgetClass, selFile, arglist, i);
  
  i = 0;
  XtSetArg(arglist[i], XtNlabel, prompt);				i++;
  XtSetArg(arglist[i], XtNresizable, True);			i++;
  XtSetArg(arglist[i], XtNtop, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNbottom, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNleft, XtChainLeft);			i++;
  XtSetArg(arglist[i], XtNright, XtChainLeft);			i++;
  XtSetArg(arglist[i], XtNborderWidth, 0);			i++;
  selFilePrompt =
    XtCreateManagedWidget("selFilePrompt",
                          labelWidgetClass, selFileForm, arglist, i);
  
  i = 0;
  XtSetArg(arglist[i], XtNforeground, &SFfore);			i++;
  XtSetArg(arglist[i], XtNbackground, &SFback);			i++;
  XtGetValues(selFilePrompt, arglist, i);
  
  SFinitFont();
  
  SFentryWidth = SFbesideText + SFcharsPerEntry * SFcharWidth + SFbesideText;
  SFentryHeight = SFaboveAndBelowText + SFcharHeight + SFaboveAndBelowText;
  
  listWidth =
    SFlineToTextH + SFentryWidth + SFlineToTextH + 1 + scrollThickness;
  listHeight = SFlineToTextV + SFentryHeight + SFlineToTextV + 1 +
    SFlineToTextV + SFlistSize * SFentryHeight +
      SFlineToTextV + 1 + scrollThickness;
  
  SFpathScrollWidth = 3 * listWidth + 2 * listSpacing + 4;
  
  hScrollX = -1;
  hScrollY = (SFlineToTextV + SFentryHeight + SFlineToTextV + 1 +
              SFlineToTextV + SFlistSize * SFentryHeight +
              SFlineToTextV);
  SFhScrollWidth = SFlineToTextH + SFentryWidth + SFlineToTextH;
  
  vScrollX = SFlineToTextH + SFentryWidth + SFlineToTextH;
  vScrollY = SFlineToTextV + SFentryHeight + SFlineToTextV;
  SFvScrollHeight = SFlineToTextV + SFlistSize * SFentryHeight + SFlineToTextV;
  
  SFupperX = SFlineToTextH + SFentryWidth + SFlineToTextH - 1;
  SFlowerY = SFlineToTextV + SFentryHeight + SFlineToTextV + 1 + SFlineToTextV;
  SFupperY = (SFlineToTextV + SFentryHeight + SFlineToTextV + 1 +
              SFlineToTextV + SFlistSize * SFentryHeight - 1);
  
  SFtextX = SFlineToTextH + SFbesideText;
  SFtextYoffset = SFlowerY + SFaboveAndBelowText + SFcharAscent;

  SFsegs[0].x1 = 0;
  SFsegs[0].y1 = vScrollY;
  SFsegs[0].x2 = vScrollX - 1;
  SFsegs[0].y2 = vScrollY;
  SFsegs[1].x1 = vScrollX;
  SFsegs[1].y1 = 0;
  SFsegs[1].x2 = vScrollX;
  SFsegs[1].y2 = vScrollY - 1;
  
  SFcompletionSegs[0].x1 = SFcompletionSegs[0].x2 = SFlineToTextH;
  SFcompletionSegs[1].x1 = SFcompletionSegs[1].x2 =
    SFlineToTextH + SFentryWidth - 1;
  
  i = 0;
  XtSetArg(arglist[i], XtNwidth, 3 * listWidth + 2 * listSpacing + 4);
  i++;
  XtSetArg(arglist[i], XtNborderColor, SFfore);			i++;
  XtSetArg(arglist[i], XtNfromVert, selFilePrompt);		i++;
  XtSetArg(arglist[i], XtNvertDistance, 10);			i++;
  XtSetArg(arglist[i], XtNresizable, True);			i++;
  XtSetArg(arglist[i], XtNtop, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNbottom, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNleft, XtChainLeft);			i++;
  XtSetArg(arglist[i], XtNright, XtChainLeft);			i++;
  XtSetArg(arglist[i], XtNstring, SFtextBuffer);			i++;
  XtSetArg(arglist[i], XtNlength, MAXPATHLEN);			i++;
  XtSetArg(arglist[i], XtNeditType, XawtextEdit);			i++;
  XtSetArg(arglist[i], XtNwrap, XawtextWrapWord);			i++;
  XtSetArg(arglist[i], XtNresize, XawtextResizeHeight);		i++;
  XtSetArg(arglist[i], XtNuseStringInPlace, True);		i++;
  selFileField = XtCreateManagedWidget("selFileField",
                                       asciiTextWidgetClass,
                                       selFileForm, arglist, i);
  
  XtOverrideTranslations(selFileField,
                         XtParseTranslationTable(oneLineTextEditTranslations));
  XtSetKeyboardFocus(selFileForm, selFileField);
  
  i = 0;
  XtSetArg(arglist[i], XtNorientation, XtorientHorizontal);	i++;
  XtSetArg(arglist[i], XtNwidth, SFpathScrollWidth);		i++;
  XtSetArg(arglist[i], XtNheight, scrollThickness);		i++;
  XtSetArg(arglist[i], XtNborderColor, SFfore);			i++;
  XtSetArg(arglist[i], XtNfromVert, selFileField);		i++;
  XtSetArg(arglist[i], XtNvertDistance, 30);			i++;
  XtSetArg(arglist[i], XtNtop, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNbottom, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNleft, XtChainLeft);			i++;
  XtSetArg(arglist[i], XtNright, XtChainLeft);			i++;
  selFileHScroll = XtCreateManagedWidget("selFileHScroll",
                                         scrollbarWidgetClass,
                                         selFileForm, arglist, i);
  
  XtAddCallback(selFileHScroll, XtNjumpProc,
		(XtCallbackProc) SFpathSliderMovedCallback, (XtPointer) NULL);
  XtAddCallback(selFileHScroll, XtNscrollProc,
		(XtCallbackProc) SFpathAreaSelectedCallback, (XtPointer) NULL);

  i = 0;
  XtSetArg(arglist[i], XtNwidth, listWidth);			i++;
  XtSetArg(arglist[i], XtNheight, listHeight);			i++;
  XtSetArg(arglist[i], XtNborderColor, SFfore);			i++;
  XtSetArg(arglist[i], XtNfromVert, selFileHScroll);		i++;
  XtSetArg(arglist[i], XtNvertDistance, 10);			i++;
  XtSetArg(arglist[i], XtNtop, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNbottom, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNleft, XtChainLeft);			i++;
  XtSetArg(arglist[i], XtNright, XtChainLeft);			i++;
  selFileLists[0] = XtCreateManagedWidget("selFileList1",
                                          compositeWidgetClass,
                                          selFileForm, arglist, i);
  
  i = 0;
  XtSetArg(arglist[i], XtNwidth, listWidth);			i++;
  XtSetArg(arglist[i], XtNheight, listHeight);			i++;
  XtSetArg(arglist[i], XtNborderColor, SFfore);			i++;
  XtSetArg(arglist[i], XtNfromHoriz, selFileLists[0]);		i++;
  XtSetArg(arglist[i], XtNfromVert, selFileHScroll);		i++;
  XtSetArg(arglist[i], XtNhorizDistance, listSpacing);		i++;
  XtSetArg(arglist[i], XtNvertDistance, 10);			i++;
  XtSetArg(arglist[i], XtNtop, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNbottom, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNleft, XtChainLeft);			i++;
  XtSetArg(arglist[i], XtNright, XtChainLeft);			i++;
  selFileLists[1] = XtCreateManagedWidget("selFileList2",
                                          compositeWidgetClass,
                                          selFileForm, arglist, i);

  i = 0;
  XtSetArg(arglist[i], XtNwidth, listWidth);			i++;
  XtSetArg(arglist[i], XtNheight, listHeight);			i++;
  XtSetArg(arglist[i], XtNborderColor, SFfore);			i++;
  XtSetArg(arglist[i], XtNfromHoriz, selFileLists[1]);		i++;
  XtSetArg(arglist[i], XtNfromVert, selFileHScroll);		i++;
  XtSetArg(arglist[i], XtNhorizDistance, listSpacing);		i++;
  XtSetArg(arglist[i], XtNvertDistance, 10);			i++;
  XtSetArg(arglist[i], XtNtop, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNbottom, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNleft, XtChainLeft);			i++;
  XtSetArg(arglist[i], XtNright, XtChainLeft);			i++;
  selFileLists[2] = XtCreateManagedWidget("selFileList3",
                                          compositeWidgetClass,
                                          selFileForm, arglist, i);

  for (n = 0; n < 3; n++) {
    
    i = 0;
    XtSetArg(arglist[i], XtNx, vScrollX);			i++;
    XtSetArg(arglist[i], XtNy, vScrollY);			i++;
    XtSetArg(arglist[i], XtNwidth, scrollThickness);	i++;
    XtSetArg(arglist[i], XtNheight, SFvScrollHeight);	i++;
    XtSetArg(arglist[i], XtNborderColor, SFfore);		i++;
    selFileVScrolls[n] = XtCreateManagedWidget("selFileVScroll",
                                               scrollbarWidgetClass,
                                               selFileLists[n], arglist, i);
    
    XtAddCallback(selFileVScrolls[n], XtNjumpProc,
                  (XtCallbackProc) SFvFloatSliderMovedCallback, (XtPointer) n);
    XtAddCallback(selFileVScrolls[n], XtNscrollProc,
                  (XtCallbackProc) SFvAreaSelectedCallback, (XtPointer) n);
    
    i = 0;
    
    XtSetArg(arglist[i], XtNorientation, XtorientHorizontal);
    i++;
    XtSetArg(arglist[i], XtNx, hScrollX);			i++;
    XtSetArg(arglist[i], XtNy, hScrollY);			i++;
    XtSetArg(arglist[i], XtNwidth, SFhScrollWidth);		i++;
    XtSetArg(arglist[i], XtNheight, scrollThickness);	i++;
    XtSetArg(arglist[i], XtNborderColor, SFfore);		i++;
    selFileHScrolls[n] = XtCreateManagedWidget("selFileHScroll",
                                               scrollbarWidgetClass,
                                               selFileLists[n], arglist, i);
    
    XtAddCallback(selFileHScrolls[n], XtNjumpProc,
                  (XtCallbackProc) SFhSliderMovedCallback, (XtPointer) n);
    XtAddCallback(selFileHScrolls[n], XtNscrollProc,
                  (XtCallbackProc) SFhAreaSelectedCallback, (XtPointer) n);
  }
  
  i = 0;
  XtSetArg(arglist[i], XtNlabel, ok);				i++;
  XtSetArg(arglist[i], XtNresizable, True);			i++;
  XtSetArg(arglist[i], XtNcallback, SFokSelect);			i++;
  XtSetArg(arglist[i], XtNborderColor, SFfore);			i++;
  XtSetArg(arglist[i], XtNfromVert, selFileLists[0]);		i++;
  XtSetArg(arglist[i], XtNvertDistance, 30);			i++;
  XtSetArg(arglist[i], XtNtop, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNbottom, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNleft, XtChainLeft);			i++;
  XtSetArg(arglist[i], XtNright, XtChainLeft);			i++;
  selFileOK = XtCreateManagedWidget("selFileOK", commandWidgetClass,
                                    selFileForm, arglist, i);
  
  i = 0;
  XtSetArg(arglist[i], XtNlabel, cancel);			i++;
  XtSetArg(arglist[i], XtNresizable, True);			i++;
  XtSetArg(arglist[i], XtNcallback, SFcancelSelect);		i++;
  XtSetArg(arglist[i], XtNborderColor, SFfore);			i++;
  XtSetArg(arglist[i], XtNfromHoriz, selFileOK);		i++;
  XtSetArg(arglist[i], XtNfromVert, selFileLists[0]);		i++;
  XtSetArg(arglist[i], XtNhorizDistance, 30);			i++;
  XtSetArg(arglist[i], XtNvertDistance, 30);			i++;
  XtSetArg(arglist[i], XtNtop, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNbottom, XtChainTop);			i++;
  XtSetArg(arglist[i], XtNleft, XtChainLeft);			i++;
  XtSetArg(arglist[i], XtNright, XtChainLeft);			i++;
  selFileCancel = XtCreateManagedWidget("selFileCancel",
                                        commandWidgetClass,
                                        selFileForm, arglist, i);

  XtSetMappedWhenManaged(selFile, False);
  XtRealizeWidget(selFile);
  
  /* Add WM_DELETE_WINDOW protocol */
  SFwmDeleteWindow = XInternAtom(SFdisplay, "WM_DELETE_WINDOW", False);
  XSetWMProtocols(SFdisplay, XtWindow(selFile), &SFwmDeleteWindow, 1);
  
  SFcreateGC();
  
  xtermCursor = XCreateFontCursor(SFdisplay, XC_xterm);
  
  sbRightArrowCursor = XCreateFontCursor(SFdisplay, XC_sb_right_arrow);
  dotCursor = XCreateFontCursor(SFdisplay, XC_dot);
  
  XDefineCursor(SFdisplay, XtWindow(selFileForm), xtermCursor);
  XDefineCursor(SFdisplay, XtWindow(selFileField), xtermCursor);
  
  for (n = 0; n < 3; n++) {
    XDefineCursor(SFdisplay, XtWindow(selFileLists[n]),
                  sbRightArrowCursor);
  }
  XDefineCursor(SFdisplay, XtWindow(selFileOK), dotCursor);
  XDefineCursor(SFdisplay, XtWindow(selFileCancel), dotCursor);
  
  for (n = 0; n < 3; n++) {
    XtAddEventHandler(selFileLists[n], ExposureMask, True,
                      (XtEventHandler) SFexposeList, (XtPointer) n);
    XtAddEventHandler(selFileLists[n], EnterWindowMask, False,
                       (XtEventHandler)SFenterList, (XtPointer) n);
    XtAddEventHandler(selFileLists[n], LeaveWindowMask, False,
                       (XtEventHandler)SFleaveList, (XtPointer) n);
    XtAddEventHandler(selFileLists[n], PointerMotionMask, False,
                     (XtEventHandler) SFmotionList, (XtPointer) n);
    XtAddEventHandler(selFileLists[n], ButtonPressMask, False,
                      (XtEventHandler)SFbuttonPressList, (XtPointer) n);
    XtAddEventHandler(selFileLists[n], ButtonReleaseMask, False,
                     (XtEventHandler) SFbuttonReleaseList, (XtPointer) n);
  }
  
  XtAddEventHandler(selFileField, KeyPressMask, False,
                    SFmodVerifyCallback, (XtPointer) NULL);
  
  SFapp = XtWidgetToApplicationContext(selFile);
}






FILE *
XsraSelFile(Widget toplevel,
            char *prompt, char *ok, char *cancel, char *failed,
	    char *init_path, char *mode, int (*show_entry)(),
            char **name_return)
{
  static int	firstTime = 1;
  Cardinal	i;
  Arg		arglist[20];
  XEvent		event;
  FILE		*fp;
  
  if (!prompt) {
    prompt = "Pathname:";
  }
  
  if (!ok) {
    ok = "OK";
  }
  
  if (!cancel) {
    cancel = "Cancel";
  }
  
  if (firstTime) {
    firstTime = 0;
    SFdisplay = XtDisplay(toplevel);
    SFcreateWidgets(toplevel, prompt, ok, cancel);
  } else {
    i = 0;
    
    XtSetArg(arglist[i], XtNlabel, prompt); i++;
    XtSetValues(selFilePrompt, arglist, i);
    
    i = 0;
    XtSetArg(arglist[i], XtNlabel, ok); i++;
    XtSetValues(selFileOK, arglist, i);
    
    i = 0;
    XtSetArg(arglist[i], XtNlabel, cancel); i++;
    XtSetValues(selFileCancel, arglist, i);
  }
  
  SFpositionWidget(selFile);
  XtMapWidget(selFile);
  
#ifdef HAVE_GETCWD
  if (!getcwd(SFstartDir, MAXPATHLEN)) {
#else 
#   ifdef HAVE_GETWD
  if (!getwd(SFstartDir)) {
#   else
  Well, I am badly hurt: No getwd(), no getcwd()...
#   endif
#endif /* defined(SVR4) || defined(SYSV) || defined(USG) */
      
    XtAppError(SFapp, "XsraSelFile: can't get current directory");
  }
  (void) strcat(SFstartDir, "/");
  (void) strcpy(SFcurrentDir, SFstartDir);
  
  if (init_path) {
    if (init_path[0] == '/') {
      (void) strcpy(SFcurrentPath, init_path);
      if (strncmp(
                  SFcurrentPath,
                  SFstartDir,
                  strlen(SFstartDir)
                  )) {
        SFsetText(SFcurrentPath);
      } else {
        SFsetText(&(SFcurrentPath[strlen(SFstartDir)]));
      }
    } else {
      (void) strcat(strcpy(SFcurrentPath, SFstartDir),
                    init_path);
      SFsetText(&(SFcurrentPath[strlen(SFstartDir)]));
    }
  } else {
    (void) strcpy(SFcurrentPath, SFstartDir);
  }
  
  SFfunc = show_entry;
  
  SFtextChanged();
  
  XtAddGrab(selFile, True, True);
    
  SFdirModTimerId = XtAppAddTimeOut(SFapp, (unsigned long) 1000,
                                    SFdirModTimer, (XtPointer) NULL);
  
  while (1) {
    XtAppNextEvent(SFapp, &event);
    XtDispatchEvent(&event);
    switch (SFstatus) {
    case SEL_FILE_TEXT:
      SFstatus = SEL_FILE_NULL;
      SFtextChanged();
      break;
    case SEL_FILE_OK:
      *name_return = SFgetText();
      if ((fp = SFopenFile(*name_return, mode, prompt, failed))) {
        SFprepareToReturn();
        return fp;
      }
      SFstatus = SEL_FILE_NULL;
      break;
    case SEL_FILE_CANCEL:
      SFprepareToReturn();
      return NULL;
    case SEL_FILE_NULL:
      break;
    }
  }
}
