/*

              ####### #         ###   #     #   ###   ######
              #       #          #     #   #     #    #     #
              #       #          #      # #      #    #     #
              #####   #          #       #       #    ######
              #       #          #      # #      #    #   #
              #       #          #     #   #     #    #    #
              ####### #######   ###   #     #   ###   #     #

                   
                  Copyright: 1994 Petr Krysl

   Czech Technical University in Prague, Faculty of Civil Engineering,
      Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                  email: pk@power2.fsv.cvut.cz
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/* This file holds the redefinition of graphic methods for the triangle */
/* with data. */

#include "Econfig.h"
#include "Ecfg.h"
#include <stdlib.h>

#include "Eview.h"
#include "Egraphic.h"
#include "Egeomops.h"
#include "Ecolors.h"
#include "Eactset.h"

#define TRI6WD_PRIVATE_HEADER
#include "Etri6wd.h"

/* ========================================================================= */
/* PRIVATE DATA */
/* ========================================================================= */

#define STRING_BUFF_SIZE 132

static char string_rep_buffer[STRING_BUFF_SIZE];

/* ========================================================================= */
/* PUBLIC */
/* ========================================================================= */

GraphicObj *
CreateTri6WD (WCRec *points, double v1, double v2, double v3, double v4, double v5, double v6)
{
  ETri6WDWCRec *thedp;
  GraphicObj *g_p;
  static BOOLEAN first_time = YES;
  
  if (first_time) {
    EMaskArray ma = {0, 0};
    EGMethodsRec mt =     {
      Type,
      AsString,
      StoreOn,
      RetrieveFrom,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      Draw,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      ModifyAssocData,
      GetAssocData,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL
    };
    
    EGSetOverrideEntry(ma, (&mt), type_func, Type);
    EGSetOverrideEntry(ma, (&mt), string_rep_func, AsString);
    EGSetOverrideEntry(ma, (&mt), store_on_func, StoreOn);
    EGSetOverrideEntry(ma, (&mt), retrieve_from_func, RetrieveFrom);
    EGSetOverrideEntry(ma, (&mt), draw_func, Draw);
    EGSetOverrideEntry(ma, (&mt), modify_assoc_data_func, ModifyAssocData);
    EGSetOverrideEntry(ma, (&mt), get_assoc_data_func, GetAssocData);

    /* set up the methods  */
    g_p = CreateTri6 (points); /* create a temp. obj. to inherit from */
    /* method table to use */
    tri6_wd_graphic_methods = EGCopyMethodTable(g_p, ma, &mt);
    EGDeepDestroyGraphics (g_p); /* destroy the temporary object */
    first_time = NO;
  }
    
  thedp = (ETri6WDWCRec *)make_node(sizeof(ETri6WDWCRec));
  g_p = EGCreateGraphics(tri6_wd_graphic_methods, (caddrt)thedp, sizeof(ETri6WDWCRec));
  EGInvokeMethod (g_p, modify_geom_func, (g_p, points));
  thedp->tri6.attributes.spec.fill_style = FILL_SOLID;
  thedp->tri6.attributes.spec.color      = 0x0;
  thedp->tri6.attributes.spec.edge_flag  = NO;
  thedp->tri6.attributes.spec.edge_color = 0x0;
  thedp->tri6.attributes.spec.shrink     = 1.0;
  thedp->tri6.attributes.spec.have_any   = YES;
  thedp->assoc_data.val[0]  = v1;
  thedp->assoc_data.val[1]  = v2;
  thedp->assoc_data.val[2]  = v3;
  thedp->assoc_data.val[3]  = v4;
  thedp->assoc_data.val[4]  = v5;
  thedp->assoc_data.val[5]  = v6;
#if 0
  {
    int i;
    for (i = 0; i < 6; i++) { fprintf (stderr, "%g %g %g ", thedp->tri6.points[i].x, thedp->tri6.points[i].y, thedp->tri6.points[i].z); }
    fprintf (stderr, "\n");
    for (i = 0; i < 6; i++) { fprintf (stderr, "%g ", thedp->assoc_data.val[i]); }
    fprintf (stderr, "\n");
  }
#endif
  return g_p;
}



static void 
Draw(EView *v_p, GraphicObj *p)
{
  ETri6WDWCRec *thedp;

  thedp = (ETri6WDWCRec *)(p->spec.data);

  EVTri6WDPrimitive(v_p, p,
                    thedp->tri6.points,
                    thedp->assoc_data.val,
                    thedp->tri6.attributes.spec.color,
                    thedp->tri6.attributes.spec.fill_style,
                    thedp->tri6.attributes.spec.edge_flag,
                    thedp->tri6.attributes.spec.edge_color,
                    thedp->tri6.attributes.spec.shrink,
                    DRAW_MODE_DRAW);
}



static EGraphicType 
Type(GraphicObj *g_p)
{
  return EG_TRIANGLEWD3D;
}



static char *
AsString(GraphicObj *g_p)
{
  ETri6WDWCRec *l;

  l = (ETri6WDWCRec *)g_p->spec.data;
  sprintf(string_rep_buffer,
	  "TRI6WD [%f,%f,%f] [%f,%f,%f] [%f,%f,%f] [%f,%f,%f] [%f,%f,%f] [%f,%f,%f] [%f,%f,%f,%f,%f,%f]",
	  l->tri6.points[0].x, l->tri6.points[0].y, l->tri6.points[0].z,
          l->tri6.points[1].x, l->tri6.points[1].y, l->tri6.points[1].z,
          l->tri6.points[2].x, l->tri6.points[2].y, l->tri6.points[2].z,
          l->tri6.points[3].x, l->tri6.points[3].y, l->tri6.points[3].z,
          l->tri6.points[4].x, l->tri6.points[4].y, l->tri6.points[4].z,
          l->tri6.points[5].x, l->tri6.points[5].y, l->tri6.points[5].z,
          l->assoc_data.val[0],
          l->assoc_data.val[1],
          l->assoc_data.val[2],
          l->assoc_data.val[3],
          l->assoc_data.val[4],
          l->assoc_data.val[5]
          );

  return string_rep_buffer; /* Don't free this memory!!! */
}



static void 
RetrieveFrom(FILE *file_p, GraphicSpecObj *gsp)
{
  ETri6WDWCRec *p;
  int filled, edge_flag;
  char color[64];
  EPixel pixel;
  BOOLEAN success;
  double x,y,z,v;
  float shrink;
  int i;
  
  if ((p = (ETri6WDWCRec *)make_node(sizeof(ETri6WDWCRec))) != NULL) {
    for (i = 0; i < 6; i++) {
      fscanf(file_p, "%lf %lf %lf", &x, &y, &z);
      p->tri6.points[i].x = x;
      p->tri6.points[i].y = y;
      p->tri6.points[i].z = z;
    }
    for (i = 0; i < 6; i++) {
      fscanf(file_p, "%lf", &v);
      p->assoc_data.val[i] = v;
    }
    fscanf(file_p,
	   "%d %d %d %f",
	   &(p->tri6.attributes.spec.have_any),
	   &filled, &edge_flag, &shrink);
    p->tri6.attributes.spec.fill_style = filled;
    p->tri6.attributes.spec.edge_flag  = edge_flag;
    p->tri6.attributes.spec.shrink     = shrink;
    fscanf(file_p, "%s", color);
    pixel = ColorGetPixelFromString(color, &success);
    if (success)
      p->tri6.attributes.spec.color = pixel;
    fscanf(file_p, "%s", color);
    pixel = ColorGetPixelFromString(color, &success);
    if (success)
      p->tri6.attributes.spec.edge_color = pixel;
    pixel = ColorGetPixelFromString(color, &success);
  }
  gsp->data        = (caddrt)p;
  gsp->data_length = sizeof(ETri6WDWCRec);
}
 


static void 
StoreOn(FILE *file_p, GraphicObj *g_p)
{
  ETri6WDWCRec *p;
  char *color, *edge_color;
  int i;

  p = (ETri6WDWCRec *)g_p->spec.data;
  if (p->tri6.attributes.spec.have_any == YES) {
    color =
      ColorGetStringFromPixel(p->tri6.attributes.spec.color);
    if (color == NULL)
      color = "white";
    edge_color =
      ColorGetStringFromPixel(p->tri6.attributes.spec.edge_color);
    if (edge_color == NULL)
      edge_color = "white";
  } else {
    edge_color = color = "white";
  }
  for (i = 0; i < 6; i++) {
    fprintf (file_p, "%f %f %f ",
             p->tri6.points[i].x, p->tri6.points[i].y, p->tri6.points[i].z);
  }
  for (i = 0; i < 6; i++) {
    fprintf (file_p, "%f ", p->assoc_data.val[i]);
  }
    
  fprintf(file_p,
	  "\n%d %d %d %f\n%s\n%s\n",
	  (p->tri6.attributes.spec.have_any),
	  p->tri6.attributes.spec.fill_style,
	  p->tri6.attributes.spec.edge_flag,
	  p->tri6.attributes.spec.shrink,
	  color, edge_color);
}


static GraphicObj *
ModifyAssocData(GraphicObj *g_p, FPNum *vals)
{
  ETri6WDWCRec *thedp = (ETri6WDWCRec *)g_p->spec.data;
  int i;
  for (i = 0; i < 6; i++) {  thedp->assoc_data.val[i]  = vals[i]; }
#if 0
  {
    int i;
    for (i = 0; i < 6; i++) { fprintf (stderr, "%g ", thedp->assoc_data.val[i]); }
    fprintf (stderr, "\n");
  }
#endif
  return g_p;
}



static BOOLEAN
GetAssocData(GraphicObj *g_p, FPNum *val, int *nvals)
{
  ETri6WDWCRec *p = (ETri6WDWCRec *)g_p->spec.data;
  int i;
  for (i = 0; i < 6; i++) {  val[i] = p->assoc_data.val[i]; }
  *nvals  = 6;

  return YES; /* the values are valid */
}

